import os
import pandas as pd
from pyadlml.constants import ACTIVITY, VALUE, START_TIME, END_TIME, TIME, NAME, DEVICE
from pyadlml.dataset.io.downloader import WebZipDownloader
from pyadlml.dataset.io.remote import DataFetcher


def fetch_casas(testbed='aruba', keep_original=True, cache=True, retain_corrections=False,
                      folder_path=None) -> dict:
    """
    Fetches one of CASAS datasets from the internet. The original dataset or its 
    cached version is stored in the data_home folder.

    Parameters
    ----------
    testbed : str one of ['aruba', 'cairo', 'kyoto_2010', 'milan', 'tulum'], default='aruba'
        Determines which dataset is loaded.
    keep_original : bool, default=True
        Determines whether the original dataset is deleted after downloading
        or kept on the hard drive.
    cache : bool, default=True
        Determines whether the data object should be stored as a binary file for quicker access.
        For more information how caching is used refer to the :ref:`user guide <storage>`.
    retain_corrections : bool, default=False
        When set to *true* data points that are changed or dropped during preprocessing
        are listed in the respective attributes of the data object.  Fore more information
        about the attributes refer to the :ref:`user guide <error_correction>`.

    Examples
    --------
    >>> from pyadlml.dataset import fetch_casas
    >>> data = fetch_casas(house='milan')

    Returns
    -------
    data : dict
        A dictionary containg the activity and device dataframe
    """
    assert testbed in ['milan', 'cairo', 'kyoto_2010', 'tulum', 'aruba']
    if testbed == 'milan':
        return CasasMilanFetcher()(keep_original=keep_original, cache=cache, #load_cleaned=load_cleaned, retain_corrections=retain_corrections, folder_path=folder_path
        )

    elif testbed == 'cairo':
        return CasasCairoFetcher()(keep_original=keep_original, cache=cache, #load_cleaned=load_cleaned,
                                retain_corrections=retain_corrections, folder_path=folder_path
        )

    elif testbed == 'tulum':
        return CasasTulumFetcher()(keep_original=keep_original, cache=cache, #load_cleaned=load_cleaned,
                                retain_corrections=retain_corrections, folder_path=folder_path
        )

    elif testbed == 'kyoto_2010':

        return CasasKyotoFetcher()(keep_original=keep_original, cache=cache, #load_cleaned=load_cleaned,
                                retain_corrections=retain_corrections, folder_path=folder_path
        )

    elif testbed == 'aruba':
        return CasasArubaFetcher()(keep_original=keep_original, cache=cache, #load_cleaned=load_cleaned,
                                retain_corrections=retain_corrections, folder_path=folder_path
        )


CASAS_ARUBA_URL_MEGA = 'https://mega.nz/file/QA5hEToD#V0ypxFsxiwWgVV49OzhsX8RnMNTX8MYSUM2TLL1xX6w'
CASAS_ARUBA_FILENAME = 'casas_aruba.zip'
CASAS_ARUBA_URL = "http://casas.wsu.edu/datasets/aruba.zip"

class CasasArubaFetcher(DataFetcher):
    def __init__(self, *args, **kwargs):

        downloader = WebZipDownloader(
            url=CASAS_ARUBA_URL,
            dataset_name='casas_aruba'
        )

        super().__init__(
            dataset_name='casas_aruba',
            downloader=downloader,
            correct_activities=True,
            correct_devices=True
        )

    def _fix_data(self, path, fp_corr):
        """
        as the data is very inconsistent with tabs and spaces this is to make it alright again
        produces: 
            date time,id,value,activity 
        """
        data_path_tf = fp_corr
        with open(path, 'r') as f_o, open(data_path_tf, 'w') as f_t:
            i= 1
            for line in f_o.readlines():
                s = line.split()
                # weird enterhome and leave home devices that appear inconsistently and are omitted
                if s[2] in ['ENTERHOME', 'LEAVEHOME']:
                    continue
                # there is an error in line 1476694 where M014 is replaced with a 'c'
                if s[2] == 'c':
                    s[2] = 'M014'
                # 'c' and '5' are randomly added onto values - remove them
                if 'c' in s[3]:
                    s[3] = s[3].replace('c', '')                
                if '5' in s[3] and s[2][0] == 'M':
                    s[3] = s[3].replace('5', '')
                if s[3] in ['ONM026', 'ONM009', 'ONM024']:         
                    s[3] = 'ON'                
                # line 886912 error should be on
                if s[2][0] == 'M' and len(s[3]) == 1:
                    s[3] = 'ON'                
                # line 900915 error should be off 
                if s[2][0] == 'M' and len(s[3]) == 2 and s[1] == '18:13:47.291404':
                    s[3] = 'OFF'
                # add new line
                new_line = " ".join(s[:2]) + "," + ",".join(s[2:4])
                try:
                    s[4] # test if there is an activity
                    new_line += "," + " ".join(s[4:])
                except:
                    pass
                    
                f_t.write(new_line + "\n")
            f_t.close()
            f_o.close()


    def _val_activity_count(df_act):
        # confirm data assumptions from readme
        assert len(df_act[df_act[ACTIVITY] == 'Meal_Preparation']) == 1606
        # observed 2919 times line below
        len(df_act[df_act[ACTIVITY] == 'Relax'])# == 2910 # does not correspond to authors reported values
        assert len(df_act[df_act[ACTIVITY] == 'Eating']) == 257
        assert len(df_act[df_act[ACTIVITY] == 'Work'])== 171
        assert len(df_act[df_act[ACTIVITY] == 'Sleeping']) == 401
        assert len(df_act[df_act[ACTIVITY] == 'Wash_Dishes']) == 65
        assert len(df_act[df_act[ACTIVITY] == 'Bed_to_Toilet']) == 157
        assert len(df_act[df_act[ACTIVITY] == 'Enter_Home']) == 431
        assert len(df_act[df_act[ACTIVITY] == 'Leave_Home']) == 431
        assert len(df_act[df_act[ACTIVITY] == 'Housekeeping']) == 33
        assert len(df_act[df_act[ACTIVITY] == 'Respirate']) == 6


    def load_data(self, folder_path):
        fp = folder_path.joinpath("data")

        fp_corr = folder_path.joinpath('corrected_data.csv')
        self._fix_data(fp, fp_corr)

        df_dev, df_act = _load_corrected_dfs(fp_corr)

        lst_act = df_act[ACTIVITY].unique()
        lst_dev = df_dev[DEVICE].unique()

        return dict(
            activities=df_act,
            devices=df_dev,
            activity_list=lst_act,
            device_list=lst_dev
        )







CASAS_MILAN_URL = "http://casas.wsu.edu/datasets/milan.zip"

class CasasMilanFetcher(DataFetcher):
    def __init__(self, *args, **kwargs):

        downloader = WebZipDownloader(
            url=CASAS_MILAN_URL,
            dataset_name='casas_milan'
        )

        super().__init__(
            dataset_name='casas_milan',
            downloader=downloader,
            correct_activities=True,
            correct_devices=True
        )

    def _fix_data(self, path, fp_corr):
        """
        as the data is very inconsistent with tabs and spaces this is to make it alright again
        produces: 
            date time,id,value,activity 
        """
        with open(path, 'r') as f_o, open(fp_corr, 'w') as f_t:
            for i, line in enumerate(f_o.readlines()):
                s = line[:-1].split('\t')

                # one tab \t to much 
                if i == 10285:
                    s.remove('')

                # the value ON is mislabeled as ON0 
                if i == 275005:
                    s[2] = 'ON'

                # The value on of device M019 is mislabeled as O
                if i == 433139:
                    s[2] = 'ON'

                # The value on of device M022 is mislabeled as ON`
                if i == 174353:
                    s[2] = 'ON'


                new_line = ",".join(s)
                try:
                    s[4] # test if there is an activity
                    new_line += "," + " ".join(s[4:])
                except IndexError as e:
                    pass

                assert len(s) in [3, 4]
                    
                f_t.write(new_line + "\n")
            f_t.close()
            f_o.close()


    def _val_activity_count(df_act):
        # confirm data assumptions from readme
        assert len(df_act[df_act[ACTIVITY] == 'Meal_Preparation']) == 1606
        # observed 2919 times line below
        len(df_act[df_act[ACTIVITY] == 'Relax'])# == 2910 # does not correspond to authors reported values
        assert len(df_act[df_act[ACTIVITY] == 'Eating']) == 257
        assert len(df_act[df_act[ACTIVITY] == 'Work'])== 171
        assert len(df_act[df_act[ACTIVITY] == 'Sleeping']) == 401
        assert len(df_act[df_act[ACTIVITY] == 'Wash_Dishes']) == 65
        assert len(df_act[df_act[ACTIVITY] == 'Bed_to_Toilet']) == 157
        assert len(df_act[df_act[ACTIVITY] == 'Enter_Home']) == 431
        assert len(df_act[df_act[ACTIVITY] == 'Leave_Home']) == 431
        assert len(df_act[df_act[ACTIVITY] == 'Housekeeping']) == 33
        assert len(df_act[df_act[ACTIVITY] == 'Respirate']) == 6


    def load_data(self, folder_path):
        fp = folder_path.joinpath("data")

        fp_corr = folder_path.joinpath('corrected_data.csv')
        self._fix_data(fp, fp_corr)

        df = pd.read_csv(fp_corr,
                        sep=",",
                        #parse_dates=True,
                        infer_datetime_format=True,
                        na_values=True,
                        names=[START_TIME, 'id', VALUE, ACTIVITY],
                        engine='python'  #to ignore warning for fallback to python engine because skipfooter
                        #dtyp
                        )
        df[START_TIME] = pd.to_datetime(df[START_TIME])
        df = df.sort_values(by=START_TIME).reset_index(drop=True)

        df_dev = _get_devices_df(df)
        df_act = _get_activity_df(df)

        lst_act = df_act[ACTIVITY].unique()
        lst_dev = df_dev[DEVICE].unique()

        return dict(
            activities=df_act,
            devices=df_dev,
            activity_list=lst_act,
            device_list=lst_dev
        )





CASAS_KYOTO_URL = "http://casas.wsu.edu/datasets/twor.2010.zip"

class CasasKyotoFetcher(DataFetcher):
    def __init__(self, *args, **kwargs):

        downloader = WebZipDownloader(
            url=CASAS_KYOTO_URL,
            dataset_name='casas_kyoto_2010'
        )

        super().__init__(
            dataset_name='casas_kyoto_2010',
            downloader=downloader,
            correct_activities=True,
            correct_devices=True
        )

    def _fix_data(self, path, fp_corr):
        """
        as the data is very inconsistent with tabs and spaces this is to make it alright again
        produces: 
            date time,id,value,activity 
        """
        with open(path, 'r') as f_o, open(fp_corr, 'w') as f_t:
            delimiter = ';'
            for i, line in enumerate(f_o.readlines()):

                # Seperate with tabs and whitespaces and remove empty sets
                s = [sub.split(' ') for sub in line[:-1].split('\t')]
                s = [subsub for sub in s for subsub in sub]
                s = [item for item in s if item != '']

                if not s:
                    # the case for empty lines
                    continue

                # Join timestamp
                s = [' '.join([s[0], s[1]])] + s[2:]

                try:
                    s = self._fix_line(s, i)
                except ValueError:
                    continue

                new_line = delimiter.join(s[:3])

                try:
                    s[4] # test if there is an activity
                    new_line += delimiter + " ".join(s[3:])
                except IndexError as e:
                    pass

                f_t.write(new_line + "\n")
            f_t.close()
            f_o.close()


    def _fix_line(self, s, i):

        if i == 2082109:
            # Activity work end was not begun in the first place -> remove Activity 
            s = s[:-2]
        if i == 2082361:
            # Activity work begins but never ends -> remove Activity
            s = s[:-2]

        return s



    def _val_activity_count(df_act):
        # confirm data assumptions from readme
        assert len(df_act[df_act[ACTIVITY] == 'Meal_Preparation']) == 1606
        # observed 2919 times line below
        len(df_act[df_act[ACTIVITY] == 'Relax'])# == 2910 # does not correspond to authors reported values
        assert len(df_act[df_act[ACTIVITY] == 'Eating']) == 257
        assert len(df_act[df_act[ACTIVITY] == 'Work'])== 171
        assert len(df_act[df_act[ACTIVITY] == 'Sleeping']) == 401
        assert len(df_act[df_act[ACTIVITY] == 'Wash_Dishes']) == 65
        assert len(df_act[df_act[ACTIVITY] == 'Bed_to_Toilet']) == 157
        assert len(df_act[df_act[ACTIVITY] == 'Enter_Home']) == 431
        assert len(df_act[df_act[ACTIVITY] == 'Leave_Home']) == 431
        assert len(df_act[df_act[ACTIVITY] == 'Housekeeping']) == 33
        assert len(df_act[df_act[ACTIVITY] == 'Respirate']) == 6


    def load_data(self, folder_path):
        fp = folder_path.joinpath("data")

        fp_corr = folder_path.joinpath('corrected_data.csv')
        self._fix_data(fp, fp_corr)
        df_dev, df_act = _load_corrected_dfs(fp_corr, delimiter=';')
        from pyadlml.dataset._core.activities import ActivityDict
        lst_act_res1 = [
            'R1_Wandering_in_room',
            'R1_Sleep',
            'R1_Bed_Toilet_Transition',
            'R1_Personal_Hygiene',
            'R1_Bathing',
            'R1_Work',
            'R1_Meal_Preparation',
            'R1_Leave_Home',
            'R1_Enter_Home',
            'R1_Eating',
            'R1_Watch_TV',
            'R1_Housekeeping',
            'R1_Sleeping_Not_in_Bed'
        ]
        lst_act_res2 = [
            'R2_Wandering_in_room',
            'R2_Meal_Preparation',
            'R2_Eating',
            'R2_Work',
            'R2_Bathing',
            'R2_Leave_Home',
            'R2_Watch_TV',
            'R2_Bed_Toilet_Transition',
            'R2_Enter_Home',
            'R2_Sleep',
            'R2_Personal_Hygiene',
            'R2_Sleeping_Not_in_Bed'
        ]
        dct_act = ActivityDict({
            'resident_1': df_act[df_act[ACTIVITY].isin(lst_act_res1)],
            'resident_2': df_act[df_act[ACTIVITY].isin(lst_act_res2)],
            })

        lst_act = df_act[ACTIVITY].unique()
        lst_dev = df_dev[DEVICE].unique()


        lst_act = df_act[ACTIVITY].unique()
        lst_dev = df_dev[DEVICE].unique()

        return dict(
            activities=dct_act,
            devices=df_dev,
            activity_list=lst_act,
            device_list=lst_dev
        )






CASAS_CAIRO_URL = "http://casas.wsu.edu/datasets/cairo.zip"

class CasasCairoFetcher(DataFetcher):
    def __init__(self, *args, **kwargs):

        downloader = WebZipDownloader(
            url=CASAS_CAIRO_URL,
            dataset_name='casas_cairo'
        )

        super().__init__(
            dataset_name='casas_cairo',
            downloader=downloader,
            correct_activities=True,
            correct_devices=True
        )

    def _fix_data(self, path, fp_corr):
        """
        as the data is very inconsistent with tabs and spaces this is to make it alright again
        produces: 
            date time,id,value,activity 
        """
        with open(path, 'r') as f_o, open(fp_corr, 'w') as f_t:
            for i, line in enumerate(f_o.readlines()):
                s = line[:-1].split('\t')

                new_line = ",".join(s)

                try:
                    s[4] # test if there is an activity
                    new_line += "," + " ".join(s[4:])
                except IndexError as e:
                    pass

                assert len(s) in [3, 4]
                    
                f_t.write(new_line + "\n")
            f_t.close()
            f_o.close()


    def _val_activity_count(df_act):
        # confirm data assumptions from readme
        assert len(df_act[df_act[ACTIVITY] == 'Meal_Preparation']) == 1606
        # observed 2919 times line below
        len(df_act[df_act[ACTIVITY] == 'Relax'])# == 2910 # does not correspond to authors reported values
        assert len(df_act[df_act[ACTIVITY] == 'Eating']) == 257
        assert len(df_act[df_act[ACTIVITY] == 'Work'])== 171
        assert len(df_act[df_act[ACTIVITY] == 'Sleeping']) == 401
        assert len(df_act[df_act[ACTIVITY] == 'Wash_Dishes']) == 65
        assert len(df_act[df_act[ACTIVITY] == 'Bed_to_Toilet']) == 157
        assert len(df_act[df_act[ACTIVITY] == 'Enter_Home']) == 431
        assert len(df_act[df_act[ACTIVITY] == 'Leave_Home']) == 431
        assert len(df_act[df_act[ACTIVITY] == 'Housekeeping']) == 33
        assert len(df_act[df_act[ACTIVITY] == 'Respirate']) == 6


    def load_data(self, folder_path):
        fp = folder_path.joinpath("data")

        fp_corr = folder_path.joinpath('corrected_data.csv')
        self._fix_data(fp, fp_corr)

        df = pd.read_csv(fp_corr,
                        sep=",",
                        #parse_dates=True,
                        infer_datetime_format=True,
                        na_values=True,
                        names=[START_TIME, 'id', VALUE, ACTIVITY],
                        engine='python'  #to ignore warning for fallback to python engine because skipfooter
                        #dtyp
                        )
        df[START_TIME] = pd.to_datetime(df[START_TIME])
        df_dev = _get_devices_df(df)
        df_act = _get_activity_df(df)
        from pyadlml.dataset.util import ActivityDict

        lst_act_res1 = ['R1 sleep', 'R1 work in office', 'R1 wake']
        lst_act_res2 = ['R2 wake', 'R2 take medicine', 'R2 sleep']
        joint = [
            'Night wandering', 'Bed to toilet', 'Breakfast', 'Leave home', 
            'Lunch', 'Dinner', 'Laundry'
        ]
        dct_act = ActivityDict({
            'resident_1': df_act[df_act[ACTIVITY].isin(lst_act_res1 + joint)],
            'resident_2': df_act[df_act[ACTIVITY].isin(lst_act_res2 + joint)],
            })

        lst_act = df_act[ACTIVITY].unique()
        lst_dev = df_dev[DEVICE].unique()

        return dict(
            activities=dct_act,
            devices=df_dev,
            activity_list=lst_act,
            device_list=lst_dev
        )

CASAS_TULUM_URL = "http://casas.wsu.edu/datasets/tulum2.zip"

class CasasTulumFetcher(DataFetcher):
    def __init__(self, *args, **kwargs):

        downloader = WebZipDownloader(
            url=CASAS_TULUM_URL,
            dataset_name='tulum2010'
        )

        super().__init__(
            dataset_name='casas_tulum',
            downloader=downloader,
            correct_activities=True,
            correct_devices=True
        )

    def _fix_data(self, path, fp_corr):
        """
        as the data is very inconsistent with tabs and spaces this is to make it alright again
        produces: 
            date time,id,value,activity 
        """
        with open(path, 'r') as f_o, open(fp_corr, 'w') as f_t:
            for i, line in enumerate(f_o.readlines()):

                # Seperate with tabs and whitespaces
                s = [sub.split(' ') for sub in line[:-1].split('\t')]
                s = [subsub for sub in s for subsub in sub]

                # Join timestamp
                s = [' '.join([s[0], s[1]])] + s[2:]


                if s[-1] == 'begub':
                    # Typo in begin
                    s[-1] = 'begin'

                if 564709 == i: 
                    # Two times begin in one line 
                    s = s[:-1]
                if 575270 == i:
                    # double space in activity "Work_Table  end"
                    s.remove('')
                if i == 439036:
                    # Wrong date of bathing end
                    s[0] = '2009-11-19 00:33:14.032411' 
                
                if i == 755255:
                    # Activity Work bedroom 2 started and not ended
                    s = s[:-2]

                # Switched timestamps of Enter_Home begin and end
                if i == 898661:
                    s[0] = '2010-02-20 19:11:12.008747'
                if i == 898662:
                    s[0] = '2010-02-20 19:11:12.083821'

                # Switched timestamps of Enter_Home begin and end
                if i == 934659:
                    s[0] = '2010-02-25 19:53:48.008541'
                if i == 934660:
                    s[0] = '2010-02-25 19:53:48.081764'

                if i == 1066021:
                    s[0] = '2010-03-15 20:01:37.005982'
                if i == 1066023:
                    s[0] = '2010-03-15 20:01:37.056323'

                # Switched activities of Leave home begin and end
                if i == 1068097:
                    s[-1] = 'begin'
                if i == 1068098:
                    s[-1] = 'end'

                # Switched timestamps of Enter_Home begin and end
                if i == 1078887:
                    s[0] = '2010-03-19 18:39:31.006645'
                if i == 1078888:
                    s[0] = '2010-03-19 18:39:31.022917'

                new_line = ",".join(s[:3])
                try:
                    s[4] # test if there is an activity
                    new_line += "," + " ".join(s[3:])
                except IndexError as e:
                    pass

                f_t.write(new_line + "\n")
            f_t.close()
            f_o.close()


        """
        
                                    start_time                   end_time    activity                     diff
        3461  2009-11-19 00:29:33.003653 2009-11-19 00:28:14.032411     Bathing -1 days +23:58:41.028758
        10068 2010-02-20 19:11:12.083821 2010-02-20 19:11:12.008747  Enter_Home -1 days +23:59:59.924926
        10522 2010-02-25 19:53:48.081764 2010-02-25 19:53:48.008541  Enter_Home -1 days +23:59:59.926777
        12345 2010-03-15 20:01:37.056323 2010-03-15 20:01:37.005982  Enter_Home -1 days +23:59:59.949659
        12377 2010-03-18 12:43:24.011637 2010-03-18 12:43:21.096494  Leave_Home -1 days +23:59:57.084857
        12395 2010-03-19 18:39:31.022917 2010-03-19 18:39:31.006645  Enter_Home -1 days +23:59:59.98372 
            
        """

    def _val_activity_count(df_act):
        # confirm data assumptions from readme
        assert len(df_act[df_act[ACTIVITY] == 'Meal_Preparation']) == 1606
        # observed 2919 times line below
        len(df_act[df_act[ACTIVITY] == 'Relax'])# == 2910 # does not correspond to authors reported values
        assert len(df_act[df_act[ACTIVITY] == 'Eating']) == 257
        assert len(df_act[df_act[ACTIVITY] == 'Work'])== 171
        assert len(df_act[df_act[ACTIVITY] == 'Sleeping']) == 401
        assert len(df_act[df_act[ACTIVITY] == 'Wash_Dishes']) == 65
        assert len(df_act[df_act[ACTIVITY] == 'Bed_to_Toilet']) == 157
        assert len(df_act[df_act[ACTIVITY] == 'Enter_Home']) == 431
        assert len(df_act[df_act[ACTIVITY] == 'Leave_Home']) == 431
        assert len(df_act[df_act[ACTIVITY] == 'Housekeeping']) == 33
        assert len(df_act[df_act[ACTIVITY] == 'Respirate']) == 6


    def load_data(self, folder_path):
        fp = folder_path.joinpath("data")

        fp_corr = folder_path.joinpath('corrected_data.csv')
        self._fix_data(fp, fp_corr)

        df_dev, df_act = _load_corrected_dfs(fp_corr)
        from pyadlml.dataset.util import ActivityDict

        lst_act_res1 = [
            'R1_Sleeping_in_Bed',
        ]
        lst_act_res2 = [
            'R2_Sleeping_in_Bed',
        ]
        joint = [
            'Personal_Hygiene',
            'Bathing',
            'Leave_Home',
            'Enter_Home',
            'Meal_Preparation',
            'Watch_TV',
            'Eating',
            'Bed_Toilet_Transition',
            'Work_Table',
            'Work_Bedroom_2',
            'Yoga',
            'Wash_Dishes',
            'Work_LivingRm',
            'Work_Bedroom_1'
        ]
        dct_act = ActivityDict({
            'resident_1': df_act[df_act[ACTIVITY].isin(lst_act_res1 + joint)],
            'resident_2': df_act[df_act[ACTIVITY].isin(lst_act_res2 + joint)],
            })

        lst_act = df_act[ACTIVITY].unique()
        lst_dev = df_dev[DEVICE].unique()

        return dict(
            activities=dct_act,
            devices=df_dev,
            activity_list=lst_act,
            device_list=lst_dev
        )

def _load_corrected_dfs(fp_corr, delimiter=','):
    df = pd.read_csv(fp_corr,
                    sep=delimiter,
                    #parse_dates=True,
                    infer_datetime_format=True,
                    na_values=True,
                    names=[START_TIME, 'id', VALUE, ACTIVITY],
                    engine='python'  #to ignore warning for fallback to python engine because skipfooter
                    #dtyp
                    )
    df[START_TIME] = pd.to_datetime(df[START_TIME])
    df = df.sort_values(by=START_TIME)\
           .drop_duplicates()
    # Drop when a device is na 
    df = df[~df.iloc[:, :3].isna().any(axis=1)].reset_index(drop=True)
    df_dev = _get_devices_df(df)
    df_act = _get_activity_df(df)
    return df_dev, df_act


def _get_devices_df(df):
    df = df.copy().drop(ACTIVITY, axis=1)
    bin_mask = (df[VALUE] == 'ON') | (df[VALUE] == 'OFF')

    # preprocess only binary devices to ON-OFF--> False True
    df_binary = df[bin_mask]
    df_binary[VALUE] = (df_binary[VALUE] == 'ON')

    # preprocess only numeric devices
    num_mask = pd.to_numeric(df[VALUE], errors='coerce').notnull()
    df_num = df[num_mask]
    df_num[VALUE] = df_num[VALUE].astype(float)

    # preprocess categorical devices
    df_cat = df[~num_mask & ~bin_mask]

    # join datasets
    df = pd.concat([df_cat, df_binary, df_num], axis=0, ignore_index=True)
    df.columns = [TIME, DEVICE, VALUE]
    df = df.sort_values(by=TIME).reset_index(drop=True)

    return df


def _get_activity_df(df):
    # get all rows containing activities
    df = df.copy()[~df[ACTIVITY].isnull()][[START_TIME, ACTIVITY]]
    df[ACTIVITY] = df[ACTIVITY].astype(str).apply(lambda x: x.strip())

    act_list = list(df[ACTIVITY].unique())
    act_list.sort()
    
    new_df_lst = []
    for i in range(1, len(act_list), 2):
        activity = ' '.join(act_list[i].split(' ')[:-1])
        act_begin = act_list[i-1]
        act_end = act_list[i]
        assert activity in act_begin and activity in act_end
           
        # create subsets for begin and end of chosen activity
        df_res = df[df[ACTIVITY] == act_begin].reset_index(drop=True)
        df_end = df[df[ACTIVITY] == act_end].reset_index(drop=True)
        #assert len(df_res) == len(df_end)
        
        # append sorted end_time to start_time as they should be
        # pairwise together
        df_res[ACTIVITY] = activity
        df_res[END_TIME] = df_end[START_TIME]
        new_df_lst.append(df_res)
    
    # data preparation
    res = pd.concat(new_df_lst)
    res = res.reindex(columns=[START_TIME, END_TIME, ACTIVITY])
    res = res.sort_values(START_TIME)
    res = res.reset_index(drop=True)
    return res



import argparse
import dash_bootstrap_components as dbc
import dash
from dask.distributed import Client, LocalCluster

from pyadlml.constants import TIME, START_TIME, END_TIME, ACTIVITY
from pyadlml.dataset._core.activities import ActivityDict
from pyadlml.dataset.io import set_data_home
from pyadlml.dataset.plot.plotly.dashboard.dashboard import dashboard, create_callbacks
from pyadlml.dataset.util import fetch_by_name


"""
Example application:
 On how to use the dash-board 
 Link: http://127.0.0.1:8050/

 Add ...:/path/to/pyadlml/examples/:... to the PYTHONPATH

"""



if __name__ == '__main__':
    parser = argparse.ArgumentParser()
    parser.add_argument('-d', '--dataset', type=str, default='amsterdam')
    parser.add_argument('-i', '--identifier', type=str)
    parser.add_argument('-p', '--port', type=str, default='8050')
    parser.add_argument('-dh', '--data-home', type=str, default='/tmp/pyadlml/')
    args = parser.parse_args()

    # Setup dask
    cluster = LocalCluster(n_workers=12)
    client = Client(cluster)

    # Setup data
    set_data_home(args.data_home)
    data = fetch_by_name(args.dataset, identifier=args.identifier, cache=False)

    # Create global variables
    dct_acts = ActivityDict.wrap(data['activities'])
    df_devs = data['devices']

    dev2area = data.pop('dev2area', None)

    # Determine plot dimensions based on #devices and #activities
    start_time = min(df_devs[TIME].iloc[0], *[dct_acts[key][START_TIME].iloc[0] for key in dct_acts.keys()])
    end_time = max(df_devs[TIME].iloc[-1], *[dct_acts[key][END_TIME].iloc[-1] for key in dct_acts.keys()])
    start_time = start_time.floor('D')
    end_time = end_time.ceil('D')

    # Initialize graph and functions
    app = dash.Dash(__name__, external_stylesheets=[dbc.themes.BOOTSTRAP])
    dashboard(app, name=args.dataset, embedded=False, df_acts=dct_acts, df_devs=df_devs, 
              dev2area=dev2area, start_time=start_time, end_time=end_time)



    print(f'Start server under: http://127.0.0.1:{args.port}/')
    app.run(debug=False, host='127.0.0.1', port=args.port)



import os
from pathlib import Path
import pandas as pd
from pyadlml.constants import ACTIVITY, DEVICE, END_TIME, START_TIME
from pyadlml.dataset._core.activities import ActivityDict, correct_activities
from pyadlml.dataset._core.devices import device_boolean_on_states_to_events
from pyadlml.dataset.io.downloader import MegaDownloader
from pyadlml.dataset.io.remote import DataFetcher

UCI_ADL_BINARY_URL = 'https://mega.nz/file/AQIgDQJD#oximAQFjexTKwNP3WYzlPnOGew06YSQ2ef85vvWGN94'
UCI_ADL_BINARY_FILENAME = 'uci_adl_binary.zip'
ORD_A = 'OrdonezA'
ORD_B = 'OrdonezB'

UCI_ADL_BINARY_ORD_A_CLEANED_URL = 'https://mega.nz/file/8JoEmYwZ#gKMQa0s2P-W8d_fO2qO_KFic4S2XxGPhiGj9TfMjVs0'
UCI_ADL_BINARY_ORD_A_CLEANED_FN = 'cleaned_uci_adl_binary_A.joblib'

UCI_ADL_BINARY_ORD_B_CLEANED_URL = 'https://mega.nz/file/hBpQzYjC#Gf3JPOifb2-tpZswUAsTBoHp_7nKH0TY1RvJAPdt4BE'
UCI_ADL_BINARY_ORD_B_CLEANED_FN = 'cleaned_uci_adl_binary_B.joblib'

class UCIFetcher(DataFetcher):
    def __init__(self, *args, **kwargs):

        downloader = MegaDownloader(
            url=UCI_ADL_BINARY_URL,
            fn=UCI_ADL_BINARY_FILENAME,
            url_cleaned={
                ORD_A:UCI_ADL_BINARY_ORD_A_CLEANED_URL,
                ORD_B:UCI_ADL_BINARY_ORD_B_CLEANED_URL
            },
            fn_cleaned={
                ORD_A:UCI_ADL_BINARY_ORD_A_CLEANED_FN,
                ORD_B:UCI_ADL_BINARY_ORD_B_CLEANED_FN
            },
        )

        super().__init__(
            dataset_name='uci_adl_binary',
            downloader=downloader,
            correct_activities=True,
            correct_devices=True
        )

    def correct_activities(self, key, df_act : pd.DataFrame, ident: str) -> pd.DataFrame:
        """ Apply manual corrections to irreparable data faults in ordonez dataset
        """
        assert ident in [ORD_A, ORD_B]
        # manual activity correction for OrdonezB
        if ident == ORD_B:
            # the activity grooming is often overlapped by sleeping
            # Since grooming is more important make it the dominant activity (opinionated)
            df_act, correction_act = correct_activities(df_act, excepts=['Grooming'], retain_corrections=True)

        elif ident == ORD_A:
            # Manually replace 3 rows where START_TIME > END_TIME in the activity files
            # 72 "2011-12-01 19:28:51  2011-12-01 16:29:59  Toileting"
            # to "2011-12-01 19:28:51  2011-19-01 16:29:59  Toileting"

            # 81 "2011-12-02 12:20:41  2011-12-01 10:20:59  Grooming"
            # to "2011-12-02 12:20:41  2011-12-01 12:20:59  Grooming"

            # 83 "2011-12-02 12:27:47  2011-12-01 11:35:49  Breakfast"
            # to "2011-12-02 12:27:47  2011-12-01 12:35:49  Breakfast"
            df_act.iat[69, 1] = pd.Timestamp('2011-12-01 19:29:59')
            df_act.iat[78, 1] = pd.Timestamp('2011-12-02 12:20:59')
            df_act.iat[80, 1] = pd.Timestamp('2011-12-02 12:35:49')

            df_act, correction_act = correct_activities(df_act, retain_corrections=True)

        return df_act, correction_act


    def load_data(self, folder_path: Path, ident: str) -> dict:
        assert ident in [ORD_A, ORD_B]

        # Fix path and Ordonez B activity file
        sub_dev_file = folder_path.joinpath('{}_Sensors.txt'.format(ident))
        if ident == ORD_B:
            fix_OrdonezB_ADLS(folder_path.joinpath('OrdonezB_ADLs.txt'))
            sub_act_file = folder_path.joinpath('{}_ADLs_corr.txt'.format(ident))
        else:
            sub_act_file = folder_path.joinpath('{}_ADLs.txt'.format(ident))

        # Load activities 
        df_act = _load_activities(sub_act_file)
        #dct_act = ActivityDict({'%s'%(ident):df_act})
        dct_act = df_act

        # Load devices
        df_dev, df_areas = _load_devices(sub_dev_file)
        df_dev = device_boolean_on_states_to_events(df_dev)

        lst_act = df_act[ACTIVITY].unique()
        lst_dev = df_dev[DEVICE].unique()

        return dict(
            activities=dct_act,
            devices=df_dev,
            activity_list=lst_act,
            device_list=lst_dev,
            device2area=df_areas
        )




def fetch_uci_adl_binary(subject='OrdonezA', keep_original=True, cache=True,
                        apply_corrections=True, load_cleaned=False,
                        retain_corrections=False, folder_path=None) -> dict:
    """

    Parameters
    ----------
    subject : str of {'OrdonezA', 'OrdonezB'}, default='OrdonezA'
        decides which dataset of the two houses is loaded.
    keep_original : bool, default=True
        Determines whether the original dataset is deleted after downloading
        or kept on the hard drive.
    cache : bool
        Determines whether the data object should be stored as a binary file for quicker access.
        For more information on how caching is used refer to the :ref:`user guide <storage>`.
    retain_corrections : bool, default=False
        When set to *true* data points that are changed or dropped during preprocessing
        are listed in the respective attributes of the data object.  Fore more information
        about the attributes refer to the :ref:`user guide <error_correction>`.
    apply_corrections : bool, default=True
        When set to *true* data points that are changed or dropped during preprocessing
        are listed in the respective attributes of the data object.  Fore more information
        about the attributes refer to the :ref:`user guide <error_correction>`.

    Returns
    -------
    data : object
    """
    return UCIFetcher()(keep_original=keep_original, cache=cache, load_cleaned=load_cleaned,
                        ident=subject, apply_corrections=apply_corrections,
                        retain_corrections=retain_corrections, folder_path=folder_path
    )


def fix_OrdonezB_ADLS(path_to_file: Path) -> None:
    """ Fixes inconsistent use of tabs for delimiter in the file

    Parameters
    ----------
    path_to_file : posixPath
        path to the file OrdonezB_ADLs.csv
    """
    path_corrected = path_to_file.parent.joinpath('OrdonezB_ADLs_corr.txt')
    
    with open(path_to_file, 'r') as f_o, open(path_corrected, 'w') as f_t:
        for i, line in enumerate(f_o.readlines()):            
            if i in [0,1]: 
                f_t.write(line)  
                continue
            s = line.split()
            assert len(s) == 5
            new_line = s[0]+' '+s[1]+'\t\t'+s[2]+' '+s[3]+'\t\t'+s[4]                        
            f_t.write(new_line + "\n")
        f_t.close()
        f_o.close()

def _load_activities(act_path):
    df_act = pd.read_csv(act_path, delimiter='\t+', skiprows=[0,1], 
                         names=[START_TIME, END_TIME, ACTIVITY], engine='python')
    df_act[START_TIME] = pd.to_datetime(df_act[START_TIME])
    df_act[END_TIME] = pd.to_datetime(df_act[END_TIME])
    return df_act

def _load_devices(dev_path):
    df_dev = pd.read_csv(dev_path, delimiter='\t+', skiprows=[0, 1], 
                         names=[START_TIME, END_TIME, 'Location', 'Type', 'Place'], 
                         engine='python')
    df_dev[DEVICE] = df_dev['Place'] + ' ' + df_dev['Location'] + ' ' + df_dev['Type']
    
    # get room mapping devices
    df_locs = df_dev.copy().groupby([DEVICE, 'Type', 'Place', 'Location']).sum()
    df_locs = df_locs.reset_index().drop([START_TIME, END_TIME], axis=1)

    df_dev = df_dev[[START_TIME, END_TIME, DEVICE]]
    df_dev[START_TIME] = pd.to_datetime(df_dev[START_TIME])
    df_dev[END_TIME] = pd.to_datetime(df_dev[END_TIME])
    return df_dev, df_locs


import sys
import pathlib
working_directory = pathlib.Path().absolute()
script_directory = pathlib.Path(__file__).parent.absolute()
sys.path.append(str(working_directory))
import unittest

from pyadlml.dataset import load_act_assist
from pyadlml.dataset import ACTIVITY, DEVICE, END_TIME, START_TIME, VAL, \
    TIME
from pyadlml.feature_extraction import extract_time_bins, extract_day_of_week

SUBJECT_ADMIN_NAME = 'admin'

class TestDirtyDataset(unittest.TestCase):
    def setUp(self):
        dataset_dir = str(script_directory) + '/datasets/dirty_dataset'
        self.data = load_act_assist(dataset_dir, subjects=[SUBJECT_ADMIN_NAME])

    def test_extract_time_bins(self):
        df = self.data.df_devices
        df_res2h = extract_time_bins(df)
        df_res6h = extract_time_bins(df, resolution='6h')
        df_res10m = extract_time_bins(df, resolution='10m')
        df_res40m = extract_time_bins(df, resolution='40m')
        df_res10s = extract_time_bins(df, resolution='10s')

        assert len(df_res2h.columns) == 12
        assert len(df_res6h.columns) == 4
        assert len(df_res10m.columns) == 144
        assert len(df_res40m.columns) == 36
        assert len(df_res10s.columns) == 8640

    def test_extract_day_of_week(self):
        df = self.data.df_devices
        df_dow = extract_day_of_week(df)




if __name__ == '__main__':
    unittest.main()


from pyadlml.model_selection import train_test_split, KFold, GridSearchCV
from pyadlml.pipeline import Pipeline
from pyadlml.preprocessing import LabelMatcher, CrossValSplitter


def evaluate(data, param_grid, pipeline_steps, n_jobs=6):
    """ High level evaluation method that builds a pipeline from gs_steps does a train
        test split with leave one day out. Then the best parameters are found using
        grid-search. The best model is initialized with best_model_steps and gets
        assigned the best_model_steps

    Parameters
    ----------

    Returns
    -------
    gscv : Cross-validation object
        Returned are the cross validation object

    cm : numpy nd.array
        confusion matrix

    cr : classification report object
        classification report

    best_model : pyadlml.Pipeline
        the best model

    """
    print('splitting data in train and test set...')
    X_train, X_test, y_train, y_test, y_pre_vals = train_test_split(
        data.df_devices,
        data.df_activities,
        return_init_states=True,
        split='leave_one_day_out')


    print('Grid Search to determine best parameters...')
    ts = KFold()
    pipe = Pipeline(pipeline_steps).train()
    gscv = GridSearchCV(
        online_train_val_split=True,
        estimator=pipe,
        param_grid=param_grid,
        scoring=['accuracy'],
        verbose=2,
        refit='accuracy',
        n_jobs=n_jobs,
        cv=ts
    )
    gscv = gscv.fit(X_train, y_train)

    print('Retrain model with best parameters...')

    best_model_steps = []
    # remove the CSV splitter for the best model
    for name, step in pipeline_steps:
        if isinstance(step, CrossValSplitter):
            continue
        best_model_steps.append((name, step))

    best_model = Pipeline(best_model_steps).set_params(**gscv.best_params_)
    best_model.fit(X_train, y_train)

    print('test best model...')
    transform_params = {'sve__dev_pre_values': y_pre_vals}
    best_model.eval()
    Xenc_test, y_true = best_model[:-1].transform(X_test, y_test, **transform_params)
    y_pred = best_model[-1].predict(Xenc_test)

    print('computing confusion matrix...')
    from sklearn.metrics import confusion_matrix, ConfusionMatrixDisplay
    cm = confusion_matrix(y_true, y_pred)

    classes = []
    for step in pipe:
        if isinstance(step, LabelMatcher):
            classes = step.wr.classes_
            break

    if classes == []:
        raise ValueError("No labelencoder found in Pipeline.")

    print('generating classification report:')
    from sklearn.metrics import classification_report
    cr = classification_report(y_true, y_pred,
                               target_names=classes)
    return gscv, cm, cr, best_model



from torch import nn


class EventEmbedding(nn.Module):
    """
    Event Embedding which is consisted with under features
        1. TypeEmbedding : normal embedding matrix
        2. PositionalEmbedding : adding positional information using sin, cos
        3. Time: TBD
        sum of all these features are output of EventEmbedding
    """

    def __init__(self, type_size, embed_size, dropout=0.1):
        """
        :param vocab_size: total vocab size
        :param embed_size: embedding size of token embedding
        :param dropout: dropout rate
        """
        super().__init__()
        # self.type = TypeEmbedding(type_size=type_size, embed_size=embed_size)
        self.type = nn.Embedding(type_size+1, embed_size)
        self.position = PositionalEmbedding(d_model=embed_size)
        self.dropout = nn.Dropout(p=dropout)
        self.embed_size = embed_size

    def forward(self, sequence):
        x = self.type(sequence) + self.position(sequence)
        return self.dropout(x)


from pathlib import Path
from dash import Dash, dash_table, dcc, html
from dash.dependencies import Input, Output, State
from scipy import signal as sc_signal
import pandas as pd
import numpy as np
import argparse
import dash
from dash.exceptions import PreventUpdate
from pyadlml.constants import ACTIVITY, DEVICE, END_TIME, START_TIME, STRFTIME_PRECISE, TIME, VALUE
from pyadlml.dataset._core.activities import read_activity_csv
from pyadlml.dataset._core.devices import _generate_signal, create_sig_and_corr, device_remove_state_matching_signal, read_device_df, write_device_df
from pyadlml.dataset._datasets.activity_assistant import write_device_map, write_devices
from pyadlml.dataset.plot.plotly.acts_and_devs import activities_and_devices
from pyadlml.dataset.plot.plotly.util import dash_get_trigger_element, dash_get_trigger_value, remove_whitespace_around_fig
from pyadlml.constants import ts2str

from pyadlml.dataset.util import df_difference, fetch_by_name
from pyadlml.dataset.plot.plotly.activities import correction
import dash_bootstrap_components as dbc
from plotly.subplots import make_subplots
import plotly.graph_objects as go

"""
for dataset options look under pyadlml.constants:
    casas_aruba
    amsterdam
    mitlab_1
    mitlab_2
    aras
    kasteren_A
    kasteren_B
    kasteren_C
    tuebingen_2019
    uci_ordonezA
    uci_ordonezB
    act_assist
"""

SEP_HIT = '#' + '-'*50

@remove_whitespace_around_fig
def plot_cc(sig_match, sig_search, auto_corr, cross_corr, t, eps_corr, cc_max = None):

    assert abs(len(auto_corr) - len(cross_corr)) < 3, 'Auto and cross correlation shouuld be roughly the same length'

    x_corr = np.arange(-int(len(auto_corr)/2), int(len(auto_corr)/2))

    fig = make_subplots(rows=3, cols=1)#, shared_xaxes=True)

    fig.add_trace(
        go.Scatter(x=np.arange(len(sig_search)), y=sig_search, name='search signal'), row=1, col=1
    )

    fig.add_trace(
        go.Scatter(x=np.arange(len(sig_match))-t, y=sig_match, name='match signal'), row=1, col=1
    )


    fig.add_trace(
        go.Scatter(
            x=x_corr,
            y=cross_corr, 
            name='cross-corr(search,  match)',
        ), row=2, col=1
    )

    fig.add_trace(
        go.Scatter(
            x=x_corr,
            y=auto_corr, 
            name='auto-corr(search)',
        ), row=2, col=1
    )

    fig.add_trace(
        go.Scatter(x=[x_corr[0], x_corr[-1]],y=[eps_corr]*2, name='threshold'),
        row=2, col=1
    )
    fig.add_trace(
        go.Scatter(x=[t], y=[cross_corr[t+int(len(cross_corr)/2)]], showlegend=False),
        row=2, col=1
    )

    if cc_max is not None:
        uniques, counts = np.unique(np.sort(cc_max), return_counts=True)
        #x_cc_max = np.cumsum(counts)
        #x_cc_max = np.insert(x_cc_max, 0, 0)
        x_cc_max = np.flip(np.cumsum(np.flip(counts)))
        x_cc_max = np.insert(x_cc_max, len(x_cc_max), 0)
        y_cc_max = np.insert(uniques, 0, 0)

        fig.add_trace(
            go.Scatter(
                x=x_cc_max, 
                y=y_cc_max,
                showlegend=False,
            ), row=3, col=1
        )
        fig.add_trace(
            go.Scatter(
                x = [0, max(x_cc_max)],
                y = [eps_corr]*2,
                showlegend=False, 
            ), row=3, col=1
        )
        fig.update_xaxes(title='lag', row=3, col=1)
        fig.update_yaxes(title='Positive predicted', row=3, col=1)
    return fig



def del_op2str(row: pd.Series):
    row_lst = f"\t[['{ts2str(row[TIME])}','{row[DEVICE]}','{row[VALUE]}']]"
    return f"idx_to_del = get_index_matching_rows(df_devs, {row_lst})\n"\
            + f"df_devs = df_devs.drop(index=idx_to_del).reset_index(drop=True)\n\n"

def add_op2str(row: pd.Series):
    row_dict = "{"\
            + f"'{TIME}': pd.Timestamp('{(row[TIME])}'), "\
            + f"'{DEVICE}': '{row[DEVICE]}'), "\
            + f"'{VALUE}': '{row[VALUE]}'"\
    + "}"
    return f"new_row=pd.Series({row_dict})\n"\
            + "print('adding: ', new_row)\n"\
            + "df_devs = pd.concat([df_devs, new_row.to_frame().T], axis=0).reset_index(drop=True)\n"


def plot_main(df_devs, df_acts, dev_name, signal_orig, zoomed_range):
    fig = make_subplots(rows=1, cols=1) # TODO refactor, shared x
    fig = activities_and_devices(df_acts=df_acts, df_devs=df_devs, states=True, 
                                zoomed_range=zoomed_range, fig=fig, row=1, col=1)
    fig.update(layout_showlegend=False)
    # TODO plot perfect signal resized and centered at maximum cc
    #fig.add_trace(

    #)
    # TODO plot signal under
    return fig

if __name__ == '__main__':
    parser = argparse.ArgumentParser()
    parser.add_argument('--device', type=str, required=True)
    parser.add_argument('-d', '--dataset', type=str, default='amsterdam')
    parser.add_argument('-i', '--identifier', type=str)
    parser.add_argument('-p', '--port', type=str, default='8050')
    parser.add_argument('-o', '--out-folder', type=str, default='/tmp/pyadlml/clean_device_signals')
    args = parser.parse_args()


    sel_dev = args.device

    out_folder = Path(args.out_folder)
    out_folder.mkdir(parents=True, exist_ok=True)

    code_str = "# Awesome script\n" \
            + "import pandas as pd\n"\
            + "from pyadlml.dataset._core.devices import get_index_matching_rows\n"

    fp_code = out_folder.joinpath(f'clean_dev_signal_{args.device}.py')
    fp_df_devs = out_folder.joinpath('devices.csv')

    if not fp_code.exists():
        with open(fp_code, mode='w') as f:
            f.write(code_str)


    # Setup data
    data = fetch_by_name(args.dataset, identifier=args.identifier, \
                         cache=False, retain_corrections=True, keep_original=True)

    # Create global variables
    df_devs = data['devices']
    df_acts = data['activities']
    write_devices(df_devs, fp_df_devs)

    assert sel_dev in df_devs[DEVICE].unique(), 'Check if device is present in dataset.'

    app = Dash(__name__, external_stylesheets=[dbc.themes.BOOTSTRAP])

    #-------------------------------------------
    # Parameters to adjust 
    search_signal = [
        (True, '4s'),
        (False, '2s'),
        (True, '1s'),
        (False, '6s')
    ]

    # The dataframe is search beforehand based on device one state (i.e. True, '1s'). Combined with a 
    # tolerance (eps_state) matches are genererated. Cross correlation is computed only around those matches.
    matching_state_idx = 2
    #-------------------------------------------


    # Tolerance around matching state to be detected as a match
    eps_state = '0.2s'
    # Tolerance for cross correlation w.r.t. auto correlation to highlight as a match
    eps_corr = 0.2
    sig_search = [(s[0], pd.Timedelta(s[1])) for s in search_signal]

    code_str += f'search_signal = {str(search_signal)}\n' 
    code_str += f'matching_state_idx = {matching_state_idx}\n' 
    code_str += f'eps_state = {eps_state}\n\n'

    # Find 
    td = pd.Timedelta(sig_search[matching_state_idx][1])
    state = sig_search[matching_state_idx][0]
    eps_state = pd.Timedelta(eps_state)

    # Create auto correlation and retrieve a proper threshold
    ss_discrete = _generate_signal(sig_search, dt='250ms')
    auto_corr = sc_signal.correlate(ss_discrete, ss_discrete, mode='full')
    auto_corr_max = auto_corr.max()

    # Compute the number of counts that the signal may deviate
    # but still counts as a match
    # For example if a window is 72 units long and the maximum
    # correlation would also be 72 for a signal with the same length. Therefore
    # the eps_corr count would be 14 if the signal is allowed to differ 20%
    eps_corr = int(len(ss_discrete)*eps_corr)

    # Count total length and length up and after matching state of signal
    ss_total_dt, ss_dt_prae_match, ss_dt_post_match = [pd.Timedelta('0s')]*3
    for i, (s, dt) in enumerate(sig_search):
        ss_total_dt += dt
        if i < matching_state_idx:
            ss_dt_prae_match += dt
        if i > matching_state_idx:
            ss_dt_post_match += dt

    plot_dt_around_sig = ss_total_dt*0.25

    def create_hit_list(df_devs):
        df = df_devs[df_devs[DEVICE] == sel_dev].copy()
        df['to_convert'] = False
        df['diff'] = pd.Timedelta('0ns')
        df['target'] = False

        df['diff'] = df[TIME].shift(-1) - df[TIME]
        df['target'] = (td - eps_state < df['diff'])\
                    & (df['diff'] < td + eps_state)\
                    & (df[VALUE] == state)

        # Correct the case where the first occurence is already a match
        df.at[df.index[-1], 'diff'] = ss_total_dt

        # Get indices of hits and select first match for display
        hits = df[(df['target'] == True)].index.to_list()
        return hits, df


    # Load device dataframe and recreate hits
    def reload_devices(code_str):
        lcls = {'df_devs': read_device_df(fp_df_devs)} 
        exec(code_str, {}, lcls)
        df_devs = lcls['df_devs']
        hits, df = create_hit_list(df_devs)
        return df_devs, df, hits


    hits, df = create_hit_list(df_devs)
    hit_idx = hits[0]
    sm_discrete, cross_corr = create_sig_and_corr(df, hit_idx, ss_dt_prae_match, ss_dt_post_match, ss_discrete)

    cc_max_lst = []
    for h in hits:
        _, cc = create_sig_and_corr(df, h, ss_dt_prae_match, ss_dt_post_match, ss_discrete)
        cc_max_lst.append(max(cc))

    ts = df_devs.loc[hit_idx, TIME]
    zr_st, zr_et = ts - plot_dt_around_sig, ts + ss_total_dt + plot_dt_around_sig

    HEADER_STR = f'Correcting {sel_dev}; Match: #%s/%s'

    app.layout = dbc.Container([
        dbc.Row([
            dcc.Store(id="curr-id"),
            dbc.Col(
                dcc.Graph(
                    id='correction',
                    figure=plot_main(df_devs, df_acts, sel_dev, sig_search, zoomed_range=[zr_st, zr_et])
                ), md=6
            ),
            dbc.Col([
                dcc.Graph(
                    id='correlation',
                    figure=plot_cc(sm_discrete, ss_discrete, auto_corr, cross_corr, t=0, eps_corr=eps_corr, cc_max=cc_max_lst)
                ),
                html.P("Lag:"),
                dcc.Slider(-len(ss_discrete), len(ss_discrete), id='cc_slider', marks=None, value=0),
                html.P("Threshold:"),
                dcc.Slider(0, auto_corr_max, id='thresh_slider', marks=None, value=eps_corr)
                ], md=6
            )
        ], className='mb-3'),
        dbc.Row([html.H6(id='header', children=HEADER_STR%(0, len(hits))),
        ], className='mb-2'),
        dbc.Row(
            dbc.Textarea(id='comment', readOnly=False, 
                        style={'height': '100px'}
                        #style={'resize': 'none', 'overflow':'hidden'}#, 'height': '100px'}
            ),
            style={'height':'100%'},
            className='mb-1'
        ),
        dbc.Row([
            dbc.ButtonGroup([
                dbc.Button(id="prev", children="Previous", color="secondary"), 
                dbc.Button(id="revert", children="revert last change", color="secondary"),  
                dbc.Button(id="next", children="Next", color="secondary"), 
            ])
        ], className='mb-2'),

    ], style={'width': 1500, 'margin': 'auto'})

    @app.callback(
        Output('correction', 'figure'),
        Output('correlation', 'figure'),
        Output('curr-id', 'data'),
        Output('header', 'children'),
        Output('comment', 'value'),
        Input('prev', 'n_clicks'),
        Input('revert', 'n_clicks'),
        Input('next', 'n_clicks'),
        Input('cc_slider', 'value'),
        Input('thresh_slider', 'value'),
        Input('correction', 'clickData'),
        State('correction', 'figure'),
        State('correlation', 'figure'),
        State('curr-id', 'data'),
        State('comment', 'value')
    )
    def display_output(prev_click, revert_click, next_click, cc_slider, t_slider,
                       click_main, fig_main, fig_corr, curr_hit, comment
        ):
        curr_hit = 0 if curr_hit is None else curr_hit


        trigger = dash_get_trigger_element()
        if trigger is None or trigger == '':
            raise PreventUpdate


        # Open code
        f = open(fp_code, mode='r')
        code_str = ''.join(f.readlines())
        f.close()

        df_devs, df, hits = reload_devices(code_str)

        cc_max_lst = []
        for h in hits:
            _, cc = create_sig_and_corr(df, h, ss_dt_prae_match, ss_dt_post_match, ss_discrete)
            cc_max_lst.append(max(cc))

            
        if trigger == 'cc_slider' or trigger == 'thresh_slider':
            hit_idx = hits[curr_hit]
            sm_discrete, cross_corr = create_sig_and_corr(df, hit_idx, ss_dt_prae_match, ss_dt_post_match, ss_discrete)
            fig_corr = plot_cc(sm_discrete, ss_discrete, auto_corr, cross_corr, t=int(cc_slider), eps_corr=t_slider, 
                               cc_max=cc_max_lst)

        elif trigger == 'correction':
            dev = click_main['points'][0]['label']
            if dev != sel_dev:
                raise PreventUpdate

            time = pd.Timestamp(click_main['points'][0]['base'])
            matches = df_devs[df_devs[TIME] == time].index
            assert len(matches) == 1

            # Create heading
            code_str += "\n" + SEP_HIT + "\n"
            code_str += f"# Hit_nr.: {curr_hit}\n"

            # Insert comment
            comment = '' if comment is None else comment
            code_str += '\n'.join(['# ' + s for s in comment.split('\n')]) + '\n\n'

            match_idx = matches[0]
            df_tmp = df_devs[df_devs[DEVICE] == sel_dev].copy().reset_index()
            df_tmp_idx = df_tmp[df_tmp['index'] == match_idx].index[0] 

            if False: #df_tmp_idx-1 < 0:
                # 1. case :|~~~e______e~~~~~    -->     Do nothing
                #            |remove             
                # 2. case :e~~~e______e~~~~~    -->     e______e~~~~~ 
                #            |remove             
                df_dev_post_idx = df_tmp.at[df_tmp_idx+1, 'index']
                code_str += del_op2str(df_devs.loc[match_idx, :])
                code_str += del_op2str(df_devs.loc[df_dev_post_idx, :])
                print('Warning: Boundary condition removal:::: check what was done!')
            elif df_tmp_idx+1 > len(df_tmp):
                # 1. case :______e~~~|          --> Do nothing 
                #            |remove             
                # 2. case :______e~~~e can not happen since clicking on ~~~ will retreive left e TODO 
                #            |remove             
                raise NotImplementedError
            else:
                df_dev_post_idx = df_tmp.at[df_tmp_idx+1, 'index']
                code_str += del_op2str(df_devs.loc[match_idx, :])
                code_str += del_op2str(df_devs.loc[df_dev_post_idx, :])



            df_devs, df, hits = reload_devices(code_str)

            range_from_fig = lambda f: [pd.Timestamp(ts) for ts in f['layout']['xaxis']['range']]
            fig_range = range_from_fig(fig_main)

            fig_main = plot_main(df_devs, df_acts, sel_dev, sig_search, zoomed_range=fig_range)

        elif trigger == 'revert':
            # Remove last action
            code_str = SEP_HIT.join(code_str.split(SEP_HIT)[:-1])

            df_devs, df, hits = reload_devices(code_str)
            range_from_fig = lambda f: [pd.Timestamp(ts) for ts in f['layout']['xaxis']['range']]
            fig_range = range_from_fig(fig_main)

            fig_main = plot_main(df_devs, df_acts, sel_dev, sig_search, zoomed_range=fig_range)


        elif trigger == 'prev' or trigger in 'next':
            if trigger == 'prev':
                curr_hit = max(0, curr_hit-1) 
            else:
                curr_hit = min(curr_hit+1, len(hits)-1)

            comment = ''

            hit_idx = hits[curr_hit]
            sm_discrete, cross_corr = create_sig_and_corr(df, hit_idx, ss_dt_prae_match, ss_dt_post_match, ss_discrete)

            # Get area around matched state and update main window
            #ts = df.loc[df.index[df_hit_idx], TIME]
            ts = df.loc[hit_idx, TIME]
            zr_st, zr_et = ts - plot_dt_around_sig, ts + ss_total_dt + plot_dt_around_sig

            fig_main = go.Figure(fig_main)
            fig_main = fig_main.update_layout(
                            xaxis=dict(
                                type='date',
                                range=[zr_st, zr_et]
                                #range=[str(zr_st), str(zr_et)])
                            ),
                            #xaxis2=dict(
                            #    type='date',
                            #    range=[zr_st, zr_et]
                            #)
            )

            # Update cross corrleation plot
            fig_corr = plot_cc(sm_discrete, ss_discrete, auto_corr, cross_corr, t=int(cc_slider), eps_corr=t_slider,
                               cc_max=cc_max_lst)

        else:
            raise ValueError

        with open(fp_code, mode='w') as f:
            f.write(code_str)


        return fig_main, fig_corr, curr_hit, HEADER_STR%(curr_hit, len(hits)), comment

    app.run_server(debug=False, port=args.port, host='127.0.0.1')



import torch
from torch import nn
from torch.nn.parameter import Parameter
import numpy as np


class FlattenConsecutive(nn.Module):
    def __init__(self, d):
        super().__init__()
        """
        Parameters
        ----------
        d : int 
            dilation factor
        """
        self.d = d

    def forward(self, x):
        B, T, E = x.shape   # Batch, Sequence length T, embedding dimension E
        try:
            x = x.view(B, T//self.d, E*self.d)
        except RuntimeError:
            x = x.contiguous().view(B, T//self.d, E*self.d)

        # Do this if dilation is uneven, 
        if x.shape[1] == 1:
            x = x.squeeze(1)

        self.out = x
        return self.out
            
    def parameters(self):
        return []
    

class Embedding(nn.Module):

    def __init__(self, num_embeddings, embedding_dim):
        super().__init__()
        self.weight = nn.Parameter(torch.randn((num_embeddings, embedding_dim)))
    
    def forward(self, IX):
        """
        Plug out the emb vectors for each dev given their indices

        (B, T) ->  (B, T, E)
        """
        self.out = self.weight[IX.long()]
        return self.out
  
    def parameters(self):
        return [self.weight]
    

class Tanh(nn.Module):
    def __call__(self, x):
        super().__init__()
        self.out = torch.tanh(x)
        return self.out

    def parameters(self):
        return []


class Attention(nn.Module):
    def __init__(self):
        pass

    def forward(self, x):
        return x


class Time2Vec(nn.Module):
    """ TODO Implememnt

        https://arxiv.org/pdf/1907.05321.pdf
        paper time2vec
    """
    def __init__(self, k):
        pass
        # TODO Implement

    
        
class RelativePositionalEncoding(nn.Module):
    def __init__(self, emb_dim):
        raise NotImplementedError
        # TODO implement

class PositionalEmbedding(nn.Module):

    def __init__(self, d_model, max_len=4096):
        super().__init__()

        # Compute the positional encodings once in log space.
        pe = torch.zeros(max_len, d_model).float()
        pe.require_grad = False

        position = torch.arange(0, max_len).float().unsqueeze(1)
        angle = (torch.arange(0, d_model, 2).float() * -(math.log(10000.0) / d_model)).exp()

        pe[:, 0::2] = torch.sin(position * angle)
        pe[:, 1::2] = torch.cos(position * angle)

        pe = pe.unsqueeze(0)
        self.register_buffer('pe', pe)

    def forward(self, x):

        aa = len(x.size())
        if aa > 1:
            length = x.size(1)
        else:
            length = x.size(0)

        return self.pe[:, :length]


    @classmethod
    def plot(cls, T=300, max_pos=None):
        D = 120
        T = 300
        
        D = int(np.ceil(D / 2) * 2)
        
        from torch import tensor
        inv_freq = 1.0 / (10000 ** (torch.arange(0, T, 2).float() / T))
        pos_x = torch.arange(T, dtype=torch.float32)
        sin_inp_x = torch.einsum("i,j->ij", pos_x, inv_freq)
        emb_x = cls.pos_encoding(sin_inp_x)

        emb = torch.zeros((T, D), dtype=torch.float32)
        #emb[:, : D] = emb_x


        import matplotlib.pyplot as plt

        fig, ax = plt.subplots()
        ax.pcolormesh(emb_x, cmap='copper')
        ax.set_xlabel('Embedding length')
        ax.set_xlim((0, D))
        ax.set_ylabel('Position $t$ in Sequence $1:T$')
        ax.set_title("Positional Encoding $M_{ij}$")
        return fig


import math
def new_gelu(x):
    """
    Implementation of the GELU activation function currently in Google BERT repo (identical to OpenAI GPT).
    Reference: Gaussian Error Linear Units (GELU) paper: https://arxiv.org/abs/1606.08415
    """
    return 0.5 * x * (1.0 + torch.tanh(math.sqrt(2.0 / math.pi) * (x + 0.044715 * torch.pow(x, 3.0))))


class LayerNorm(nn.Module):
    """ LayerNorm but with an optional bias. PyTorch doesn't support simply bias=False """

    def __init__(self, ndim, bias):
        super().__init__()
        self.weight = nn.Parameter(torch.ones(ndim))
        self.bias = nn.Parameter(torch.zeros(ndim)) if bias else None

    def forward(self, input):
        return F.layer_norm(input, self.weight.shape, self.weight, self.bias, 1e-5)


"""
The :mod:`sklearn.pipeline` module implements utilities to build a composite
estimator, as a chain of transforms and estimators.
"""
# Author: Edouard Duchesnay
#         Gael Varoquaux
#         Virgile Fritsch
#         Alexandre Gramfort
#         Lars Buitinck
# License: BSD
from joblib import Parallel
from sklearn.base import clone, TransformerMixin, BaseEstimator
from sklearn.utils.fixes import delayed
from sklearn.utils.metaestimators import if_delegate_has_method
from sklearn.utils import (
    Bunch,
    _print_elapsed_time,
)

from sklearn.utils.validation import check_memory, _deprecate_positional_args
from sklearn.pipeline import Pipeline as SklearnPipeline, _fit_one, _transform_one
import numpy as np
import pandas as pd


__all__ = ['Pipeline', 'YTransformer', 'XAndYTransformer', 'XOrYTransformer',
           'EvalOnlyWrapper', 'TrainOnlyWrapper']


class YTransformer():
    """
    a class that inherits from YTransformer signals to the pipeline
    to transform y given an optional X
    """
    pass


class XAndYTransformer():
    """
    a class that inherits from XAndYTransformer signals to the pipeline
    to transform X and y given both inputs X and y. The transformation is not
    applied if either X or y are missing.
    """
    pass


class XOrYTransformer(object):
    """ A class that inherits from XOrYTransformer signals to the pipeline
    to transform X or y given inputs X or y. The transformation is not
    applied if X and y are missing.
    """
    def x_or_y_transform(func):
        def wrapper(self, *args, **kwargs):
            X, y = func(self, *args, **kwargs)
            if X is None and y is not None:
                return y
            elif X is not None and y is None:
                return X
            elif X is not None and y is not None:
                return X, y
            else:
                raise ValueError("XorYtransformers transform has to return at least one not None value")
        return wrapper


# TODO try _BaseComposition for get and set params
class Wrapper(object):
    """ this class signals the pipeline to omit this step if the pipeline is running
        in trainmode
    """

    def __init__(self, wr):
        '''
        Wrapper constructor.
        @param obj: object to wrap
        '''

        # wrap the object
        self.wr = wr

        # copy and link every attribute into the wrapper
        #for attr in self.sub_estimator.__dict__:
        #    self._add_property(attr, self.sub_estimator)

        # link all callable methods
        for attr in dir(self.wr):
            try:
                if not attr.startswith('__') and callable(getattr(self.wr, attr))\
                        and attr not in ['get_params', 'set_params']:
                    exec('self.%s = wr.%s' % (attr, attr))
            except AttributeError:
                pass

        # create new child class for TrainAndEvalOnlyWrapper that inherits from the both
        # this is done because an isinstance than can detect both
        new_class_name = self.__class__.__name__ + self.wr.__class__.__name__
        child_class = type(new_class_name, (self.__class__, self.wr.__class__), {})
        self.__class__ = child_class

        # replace the init method
        #attr_lst = list(self.sub_estimator.__dict__.keys())
        #attr_str = ''
        #for attr in attr_lst[:-1]:
        #    attr_str += str(attr) + ', '
        #attr_str += attr_lst[-1]
        #meth = 'lambda {}:  self.sub_estimator.__init__({})'.format(attr_str, attr_str)
        #exec('self.__init__ = {}'.format(meth))


    def _add_property(self, attr_name, wrapped):
        value = getattr(wrapped, attr_name)
        # create local fget and fset functions
        # todo when attribute changes the getter method should set the outer attribute to the value of the inner
        #   attribute and return
        fget = lambda self: getattr(self, '_' + attr_name)
        # set the attribute of the wrapped also
        fset = lambda self, value: (setattr(wrapped, attr_name, value), setattr(self, '_' + attr_name, value))

        # add property to self
        setattr(self.__class__, attr_name, property(fget, fset))

        # add corresponding local variable
        setattr(self, '_' + attr_name, value)

    def set_params(self, **parameters):
        for parameter, value in parameters.items():
            setattr(self, parameter, value)
        return self

    def get_params(self, deep=True):
        """ returns the parameters of the wrapped object instead of the own"""
        out = dict()
        key = 'wr'
        out[key] = self.wr
        if deep and hasattr(self.wr, 'get_params'):
            deep_items = self.wr.get_params().items()
            out.update((key + '__' + k, val) for k, val in deep_items)
        return out


class TrainOrEvalOnlyWrapper(Wrapper):
    def __init__(self, wr):
        Wrapper.__init__(self, wr)


class TrainOnlyWrapper(Wrapper):
    def __init__(self, wr):
        Wrapper.__init__(self, wr)


class EvalOnlyWrapper(Wrapper):
    def __init__(self, wr):
        Wrapper.__init__(self, wr)


class Pipeline(SklearnPipeline):
    """
    Pipeline of transforms with a final estimator.

    Sequentially apply a list of transforms and a final estimator.
    Intermediate steps of the pipeline must be 'transforms', that is, they
    must implement fit and transform methods.
    The final estimator only needs to implement fit.
    The transformers in the pipeline can be cached using ``memory`` argument.

    The purpose of the pipeline is to assemble several steps that can be
    cross-validated together while setting different parameters.
    For this, it enables setting parameters of the various steps using their
    names and the parameter name separated by a '__', as in the example below.
    A step's estimator may be replaced entirely by setting the parameter
    with its name to another estimator, or a transformer removed by setting
    it to 'passthrough' or ``None``.

    Read more in the :ref:`User Guide <pipeline>`.

    .. versionadded:: 0.5

    Parameters
    ----------
    steps : list
        List of (name, transform) tuples (implementing fit/transform) that are
        chained, in the order in which they are chained, with the last object
        an estimator.

    memory : str or object with the joblib.Memory interface, default=None
        Used to cache the fitted transformers of the pipeline. By default,
        no caching is performed. If a string is given, it is the path to
        the caching directory. Enabling caching triggers a clone of
        the transformers before fitting. Therefore, the transformer
        instance given to the pipeline cannot be inspected
        directly. Use the attribute ``named_steps`` or ``steps`` to
        inspect estimators within the pipeline. Caching the
        transformers is advantageous when fitting is time consuming.

    verbose : bool, default=False
        If True, the time elapsed while fitting each step will be printed as it
        is completed.

    Attributes
    ----------
    named_steps : :class:`~sklearn.utils.Bunch`
        Dictionary-like object, with the following attributes.
        Read-only attribute to access any step parameter by user given name.
        Keys are step names and values are steps parameters.

    See Also
    --------
    make_pipeline : Convenience function for simplified pipeline construction.

    Examples
    --------
    >>> from sklearn.svm import SVC
    >>> from sklearn.preprocessing import StandardScaler
    >>> from sklearn.datasets import make_classification
    >>> from sklearn.model_selection import train_test_split
    >>> from sklearn.pipeline import Pipeline
    >>> X, y = make_classification(random_state=0)
    >>> X_train, X_test, y_train, y_test = train_test_split(X, y,
    ...                                                     random_state=0)
    >>> pipe = Pipeline([('scaler', StandardScaler()), ('svc', SVC())])
    >>> # The pipeline can be used as any other estimator
    >>> # and avoids leaking the test set into the train set
    >>> pipe.fit(X_train, y_train)
    Pipeline(steps=[('scaler', StandardScaler()), ('svc', SVC())])
    >>> pipe.score(X_test, y_test)
    0.88
    """

    @_deprecate_positional_args
    def __init__(self, steps, *, memory=None, verbose=False):
        self.steps = steps
        self.memory = memory
        self.verbose = verbose
        self._train = True
        self._eval = False
        self._prod = False
        self._validate_steps()

    def train(self):
        """ sets pipeline in train mode """
        self._train = True
        self._eval = False
        self._prod = False
        return self

    def eval(self):
        """ sets pipeline in eval mode """
        self._train = False
        self._eval = True
        self._prod = False
        return self

    def prod(self):
        """ sets pipeline in production mode """
        self._train = False
        self._eval = False
        self._prod = True
        return self

    def is_in_train_mode(self):
        return self._train

    def is_in_eval_mode(self):
        return self._eval

    def is_in_prod_mode(self):
        return self._prod

    def get_times(self):
        from pyadlml.preprocessing.preprocessing import FinalTimeTransformer
        for _, name, transform in self._iter():
            if isinstance(transform, FinalTimeTransformer):
                return transform.times_

        raise ValueError('No transformer saved the forwarded times. Readjust the pipeline')


    def construct_y_times_and_X(self, X_val, y_val):
        from pyadlml.preprocessing.preprocessing import FinalTimeTransformer
        from pyadlml.preprocessing.windows import Windows

        final_time_name, win_name, win_pos = None, None, -1

        for i, name, transform in self._iter():
            if isinstance(transform, FinalTimeTransformer):
                final_time_name = name
            if isinstance(transform, Windows):
                win_name = name
                win_pos = i

        assert final_time_name is not None, 'No final time transformer specified'

        y_times = self[final_time_name].times_

        if win_name is not None:
            y_times = self[win_name].construct_target_times(y_times)
            Xt_prae_win, _ = self[:win_pos-len(self)].transform(X_val, y_val)
            X_at_times = self[win_name].construct_X_at_target(Xt_prae_win)
        else:
            X_at_times, _ = self[:-1].transform(X_val, y_val)

        return y_times, X_at_times
                        


    def __getitem__(self, ind):
        """Returns a sub-pipeline or a single esimtator in the pipeline
        Indexing with an integer will return an estimator; using a slice
        returns another Pipeline instance which copies a slice of this
        Pipeline. This copy is shallow: modifying (or fitting) estimators in
        the sub-pipeline will affect the larger pipeline and vice-versa.
        However, replacing a value in `step` will not affect a copy.
        """
        if isinstance(ind, slice):
            if ind.step not in (1, None):
                raise ValueError("Pipeline slicing only supports a step of 1")
            pipe_clone = self.__class__(self.steps[ind], memory=self.memory, verbose=self.verbose)
            if self.is_in_train_mode():
                pipe_clone.train()
            elif self.is_in_eval_mode():
                pipe_clone.eval()
            elif self.is_in_prod_mode():
                pipe_clone.prod()
            return pipe_clone
        try:
            name, est = self.steps[ind]
        except TypeError:
            # Not an int, try get step by name
            return self.named_steps[ind]
        return est

    # Estimator interface
    def _fit(self, X, y=None, **fit_params_steps):
        # shallow copy of steps - this should really be steps_
        self.steps = list(self.steps)
        self._validate_steps()
        # Setup the memory
        memory = check_memory(self.memory)

        fit_transform_one_cached = memory.cache(_fit_transform_one)

        for (step_idx,
             name,
             transformer) in self._iter(with_final=False,
                                        filter_passthrough=False):
            if (transformer is None or transformer == 'passthrough'):
                with _print_elapsed_time('Pipeline',
                                         self._log_message(step_idx)):
                    continue

            if hasattr(memory, 'location'):
                # joblib >= 0.12
                if memory.location is None:
                    # we do not clone when caching is disabled to
                    # preserve backward compatibility
                    cloned_transformer = transformer
                else:
                    cloned_transformer = clone(transformer)
            elif hasattr(memory, 'cachedir'):
                # joblib < 0.11
                if memory.cachedir is None:
                    # we do not clone when caching is disabled to
                    # preserve backward compatibility
                    cloned_transformer = transformer
                else:
                    cloned_transformer = clone(transformer)
            else:
                cloned_transformer = clone(transformer)

            if not self._skip_transform(cloned_transformer):
                # if cloned_transformer.
                if isinstance(cloned_transformer, YTransformer):
                    y, fitted_transformer = fit_transform_one_cached(
                        cloned_transformer, y, X, None,
                        message_clsname='Pipeline',
                        message=self._log_message(step_idx),
                        **fit_params_steps[name])
                elif isinstance(cloned_transformer, XAndYTransformer) \
                        or isinstance(cloned_transformer, XOrYTransformer):
                    X, y, fitted_transformer = fit_transform_one_cached(
                        cloned_transformer, X, y, None,
                        message_clsname='Pipeline',
                        message=self._log_message(step_idx),
                        **fit_params_steps[name])
                else:
                    # Fit or load from cache the current transformer
                    X, fitted_transformer = fit_transform_one_cached(
                        cloned_transformer, X, y, None,
                        message_clsname='Pipeline',
                        message=self._log_message(step_idx),
                        **fit_params_steps[name])
            else:
                # do nothing if it is not trainmode and the trainonly wrapper set (true)
                fitted_transformer = cloned_transformer
            # Replace the transformer of the step with the fitted
            # transformer. This is necessary when loading the transformer
            # from the cache.
            self.steps[step_idx] = (name, fitted_transformer)
        return X, y

    def _skip_transform(self, transformer):
        """
        skip the transform step if one of the following conditions is true
        """
        trainvote = not self.is_in_train_mode() and isinstance(transformer, TrainOnlyWrapper)
        evalvote = not self.is_in_eval_mode() and isinstance(transformer, EvalOnlyWrapper)
        trainandevalvote = self.is_in_prod_mode() and isinstance(transformer, TrainOrEvalOnlyWrapper)
        return trainvote or evalvote or trainandevalvote

    def fit(self, X, y=None, **fit_params):
        """Fit the model

        Fit all the transforms one after the other and transform the
        data, then fit the transformed data using the final estimator.

        Parameters
        ----------
        X : iterable
            Training data. Must fulfill input requirements of first step of the
            pipeline.

        y : iterable, default=None
            Training targets. Must fulfill label requirements for all steps of
            the pipeline.

        **fit_params : dict of string -> object
            Parameters passed to the ``fit`` method of each step, where
            each parameter name is prefixed such that parameter ``p`` for step
            ``s`` has key ``s__p``.

        Returns
        -------
        self : Pipeline
            This estimator
        """
        fit_params_steps = self._check_fit_params(**fit_params)
        Xt, yt = self._fit(X, y, **fit_params_steps)
        with _print_elapsed_time('Pipeline',
                                 self._log_message(len(self.steps) - 1)):
            if self._final_estimator != 'passthrough':
                fit_params_last_step = fit_params_steps[self.steps[-1][0]]
                self._final_estimator.fit(Xt, yt, **fit_params_last_step)

        return self

    def fit_transform(self, X, y=None, **fit_params):
        """Fit the model and transform with the final estimator

        Fits all the transforms one after the other and transforms the
        data, then uses fit_transform on transformed data with the final
        estimator.

        Parameters
        ----------
        X : iterable
            Training data. Must fulfill input requirements of first step of the
            pipeline.

        y : iterable, default=None
            Training targets. Must fulfill label requirements for all steps of
            the pipeline.

        **fit_params : dict of string -> object
            Parameters passed to the ``fit`` method of each step, where
            each parameter name is prefixed such that parameter ``p`` for step
            ``s`` has key ``s__p``.

        Returns
        -------
        Xt : array-like of shape  (n_samples, n_transformed_features)
            Transformed samples
        """
        fit_params_steps = self._check_fit_params(**fit_params)
        Xt, yt = self._fit(X, y, **fit_params_steps)

        last_step = self._final_estimator
        with _print_elapsed_time('Pipeline',
                                 self._log_message(len(self.steps) - 1)):
            if last_step == 'passthrough':
                return Xt, yt
            fit_params_last_step = fit_params_steps[self.steps[-1][0]]
            if hasattr(last_step, 'fit_transform'):
                return last_step.fit_transform(Xt, yt, **fit_params_last_step)
            else:
                return last_step.fit(Xt, yt,
                                     **fit_params_last_step).transform(Xt)

    def _transform(self, X, y=None, retrieve_last_time=False, **transform_params_steps):
        """
        Extends transform behavior by allowing for transform params and x or y transformation
        """
        #transform_params_steps = self._check_fit_params(**transform_params)

        Xt = X
        if retrieve_last_time:
            memory = Xt.copy()

        if y is None:
            for _, name, transform in self._iter():
                Xt = transform.transform(Xt, **transform_params_steps[name])
            return Xt
        else:
            yt = y
            for _, name, transform in self._iter():
                if self._skip_transform(transform):
                    continue
                if isinstance(transform, YTransformer):
                    yt = transform.transform(yt, Xt, **transform_params_steps[name])
                elif isinstance(transform, XAndYTransformer) \
                        or isinstance(transform, XOrYTransformer):
                    Xt, yt = transform.transform(Xt, yt, **transform_params_steps[name])
                else:
                    Xt = transform.transform(Xt, **transform_params_steps[name])

                if retrieve_last_time:
                    # Check if time column is 
                    if 'time' not in Xt.columns:
                        return memory['time']
                    else:
                        memory = Xt.copy()

            return Xt, yt

    @if_delegate_has_method(delegate='_final_estimator')
    def predict(self, X, **predict_params):
        """Apply transforms to the data, and predict with the final estimator

        Parameters
        ----------
        X : iterable
            Data to predict on. Must fulfill input requirements of first step
            of the pipeline.

        **predict_params : dict of string -> object
            Parameters to the ``predict`` called at the end of all
            transformations in the pipeline. Note that while this may be
            used to return uncertainties from some models with return_std
            or return_cov, uncertainties that are generated by the
            transformations in the pipeline are not propagated to the
            final estimator.

            .. versionadded:: 0.20

        Returns
        -------
        y_pred : array-like
        """
        predict_params_steps = self._check_fit_params(**predict_params_steps)
        Xt = X
        # TODO BUG in evaluation mode the transformed X_t and Xt for pipe.transform have different ouput
        for _, name, transform in self._iter(with_final=False):
            if self._skip_transform(transform) or isinstance(transform, YTransformer) \
                    or isinstance(transform, XAndYTransformer):
                continue

            Xt = transform.transform(Xt, **predict_params_steps[name])
        return self.steps[-1][-1].predict(Xt, **predict_params)

    @if_delegate_has_method(delegate='_final_estimator')
    def predict_proba(self, X, **predict_params):
        """Apply transforms, and predict_proba of the final estimator

        Parameters
        ----------
        X : iterable
            Data to predict on. Must fulfill input requirements of first step
            of the pipeline.

        Returns
        -------
        y_proba : array-like of shape (n_samples, n_classes)
        """

        predict_proba_params_steps = self._check_fit_params(**predict_params)
        Xt = X
        for _, name, transform in self._iter(with_final=False):
            if self._skip_transform(transform) or isinstance(transform, YTransformer) \
                    or isinstance(transform, XAndYTransformer):
                continue
            elif isinstance(transform, XOrYTransformer):
                Xt, _ = transform.transform(Xt, None, **predict_proba_params_steps)
            else:
                Xt = transform.transform(Xt, **predict_proba_params_steps)
        return self.steps[-1][-1].predict_proba(Xt)


    def transform(self, X, y, **transform_params):
        """Transform the data, and apply `transform` with the final estimator.

        Call `transform` of each transformer in the pipeline. The transformed
        data are finally passed to the final estimator that calls
        `transform` method. Only valid if the final estimator
        implements `transform`.

        This also works where final estimator is `None` in which case all prior
        transformations are applied.

        Parameters
        ----------
        X : iterable
            Data to transform. Must fulfill input requirements of first step
            of the pipeline.

        Returns
        -------
        Xt : ndarray of shape (n_samples, n_transformed_features)
            Transformed data.
        """
        tparams_steps, pipe_params = self._check_transform_params(**transform_params)
        return self._transform(X, y, **pipe_params, **tparams_steps)


    def _check_transform_params(self, **fit_params):
        fit_params_steps = {name: {} for name, step in self.steps if step is not None}
        pipe_params = {}
        for pname, pval in fit_params.items():
            if "__" not in pname:
                pipe_params[pname] = pval
                continue
            step, param = pname.split("__", 1)
            fit_params_steps[step][param] = pval
        return fit_params_steps, pipe_params

    @if_delegate_has_method(delegate='_final_estimator')
    def decision_function(self, X):
        """Apply transforms, and decision_function of the final estimator

        Parameters
        ----------
        X : iterable
            Data to predict on. Must fulfill input requirements of first step
            of the pipeline.

        Returns
        -------
        y_score : array-like of shape (n_samples, n_classes)
        """
        Xt = X
        for _, name, transform in self._iter(with_final=False):
            if self._skip_transform(transform) or isinstance(transform, YTransformer) \
                    or isinstance(transform, XAndYTransformer):
                continue
            elif isinstance(transform, XOrYTransformer):
                Xt, _ = transform.transform(Xt, None)
            else:
                Xt = transform.transform(Xt)
        return self.steps[-1][-1].decision_function(Xt)

    @if_delegate_has_method(delegate='_final_estimator')
    def score_samples(self, X, **score_params):
        """Apply transforms, and score_samples of the final estimator.

        Parameters
        ----------
        X : iterable
            Data to predict on. Must fulfill input requirements of first step
            of the pipeline.

        Returns
        -------
        y_score : ndarray of shape (n_samples,)
        """
        score_params_steps = self._check_fit_params(**score_params)
        Xt = X
        for _, name, transformer in self._iter(with_final=False):
            if self._skip_transform(transformer) or isinstance(transformer, YTransformer) \
                    or isinstance(transformer, XAndYTransformer):
                continue
            elif isinstance(transformer, XOrYTransformer):
                Xt, _ = transformer.fit_transform(Xt, None, score_params[name])
            else:
                Xt = transformer.transform(Xt, **score_params[name])
        return self.steps[-1][-1].score_samples(Xt)

    @if_delegate_has_method(delegate='_final_estimator')
    def predict_log_proba(self, X):
        """Apply transforms, and predict_log_proba of the final estimator

        Parameters
        ----------
        X : iterable
            Data to predict on. Must fulfill input requirements of first step
            of the pipeline.

        Returns
        -------
        y_score : array-like of shape (n_samples, n_classes)
        """
        Xt = X
        for _, name, transform in self._iter(with_final=False):
            if self._skip_transform(transform) or isinstance(transform, YTransformer) \
                    or isinstance(transform, XAndYTransformer):
                continue
            elif isinstance(transform, XOrYTransformer):
                Xt, _ = transform.fit_transform(Xt, None)
            else:
                Xt = transform.transform(Xt)
        return self.steps[-1][-1].predict_log_proba(Xt)

    @if_delegate_has_method(delegate='_final_estimator')
    def score(self, X, y=None, sample_weight=None, **score_params):
        """Apply transforms, and score with the final estimator

        Parameters
        ----------
        X : iterable
            Data to predict on. Must fulfill input rquirements of first step
            of the pipeline.

        y : iterable, default=None
            Targets used for scoring. Must fulfill label requirements for all
            steps of the pipeline.

        sample_weight : array-like, default=None
            If not None, this argument is passed as ``sample_weight`` keyword
            argument to the ``score`` method of the final estimator.

        Returns
        -------
        score : float
        """
        Xt = X
        yt = y
        score_params_steps = self._check_fit_params(**score_params)

        for _, name, transform in self._iter(with_final=False):
            if self._skip_transform(transform):
                continue
            elif isinstance(transform, XOrYTransformer):
                Xt, yt = transform.transform(Xt, yt, **score_params_steps[name])
            elif isinstance(transform, YTransformer):
                yt = transform.transform(yt, Xt, **score_params_steps[name])
            elif isinstance(transform, XAndYTransformer):
                Xt, yt = transform.transform(Xt, yt, **score_params_steps[name])
            else:
                Xt = transform.transform(Xt, **score_params_steps[name])
        score_params = {}
        if sample_weight is not None:
            score_params['sample_weight'] = sample_weight
        return self.steps[-1][-1].score(Xt, yt)


def _fit_transform_one(transformer,
                       X,
                       y,
                       weight,
                       message_clsname='',
                       message=None,
                       **fit_params):
    """
    Fits ``transformer`` to ``X`` and ``y``. The transformed result is returned
    with the fitted transformer. If ``weight`` is not ``None``, the result will
    be multiplied by ``weight``.
    """
    with _print_elapsed_time(message_clsname, message):
        if hasattr(transformer, 'fit_transform'):
            res = transformer.fit_transform(X, y, **fit_params)
        else:
            res = transformer.fit(X, y, **fit_params).transform(X)

    if (isinstance(transformer, XAndYTransformer) or isinstance(transformer, XOrYTransformer)) and weight is None:
        return *res, transformer
    elif (isinstance(transformer, XAndYTransformer) or isinstance(transformer, XOrYTransformer)) and weight is not None:
        res = res * weight
        return *res, transformer
    elif weight is None:
        return res, transformer
    else:
        return res * weight, transformer

from sklearn.pipeline import FeatureUnion as SklearnFeatureUnion

class FeatureUnion(SklearnFeatureUnion):
    """Concatenates results of multiple transformer objects.

    This estimator applies a list of transformer objects in parallel to the
    input data, then concatenates the results. This is useful to combine
    several feature extraction mechanisms into a single transformer.

    Parameters of the transformers may be set using its name and the parameter
    name separated by a '__'. A transformer may be replaced entirely by
    setting the parameter with its name to another transformer,
    or removed by setting to 'drop'.

    Read more in the :ref:`User Guide <feature_union>`.

    .. versionadded:: 0.13

    Parameters
    ----------
    transformer_list : list of (string, transformer) tuples
        List of transformer objects to be applied to the data. The first
        half of each tuple is the name of the transformer. The tranformer can
        be 'drop' for it to be ignored.

        .. versionchanged:: 0.22
           Deprecated `None` as a transformer in favor of 'drop'.

    n_jobs : int, default=None
        Number of jobs to run in parallel.
        ``None`` means 1 unless in a :obj:`joblib.parallel_backend` context.
        ``-1`` means using all processors. See :term:`Glossary <n_jobs>`
        for more details.

        .. versionchanged:: v0.20
           `n_jobs` default changed from 1 to None

    transformer_weights : dict, default=None
        Multiplicative weights for features per transformer.
        Keys are transformer names, values the weights.
        Raises ValueError if key not present in ``transformer_list``.

    verbose : bool, default=False
        If True, the time elapsed while fitting each transformer will be
        printed as it is completed.

    See Also
    --------
    make_union : Convenience function for simplified feature union
        construction.

    Examples
    --------
    >>> from sklearn.pipeline import FeatureUnion
    >>> from sklearn.decomposition import PCA, TruncatedSVD
    >>> union = FeatureUnion([("pca", PCA(n_components=1)),
    ...                       ("svd", TruncatedSVD(n_components=2))])
    >>> X = [[0., 1., 3], [2., 2., 5]]
    >>> union.fit_transform(X)
    array([[ 1.5       ,  3.0...,  0.8...],
           [-1.5       ,  5.7..., -0.4...]])
    """

    @_deprecate_positional_args
    def __init__(self, transformer_list, *, n_jobs=None,
                 transformer_weights=None, verbose=False, skip_attr_list=[]):

        self.skip_attr_list = []
        # althouth skip attribute is passesd into the constructor, generate a new one
        for trans_tuple in transformer_list:
            name = 'skip_' + trans_tuple[0]
            setattr(self, name, False)
            self.skip_attr_list.append(name)

        super(FeatureUnion, self).__init__(transformer_list, n_jobs=n_jobs,
            transformer_weights=transformer_weights, verbose=verbose)

        # the dynamically create attributes have to be passed to the class when it is cloned
        # in cross validation. THis is a Hack that creates a New class with an appropriate init
        # methods
        # replace the init method with ones where the skip_step parameters are present
        attr_str = "self, transformer_list, *, n_jobs=None, transformer_weights=None, verbose=False, skip_attr_list=[], "
        attr_assignment_str = ""
        for i, attr in enumerate(self.skip_attr_list):
            attr_str += str(attr) + '=False'
            attr_assignment_str+= 'setattr(self, "' + str(attr) + '", ' + str(attr) + '), '
            attr_str += ", "
            if i == len(self.skip_attr_list)-1:
                attr_str = attr_str[:-2]
        meth_str  = 'lambda {}: (setattr(self, "skip_attr_list", skip_attr_list), ' \
                    '{}'\
                    'super(FeatureUnion, self).__init__(transformer_list, n_jobs=n_jobs, ' \
                    'transformer_weights=transformer_weights, verbose=verbose),' \
                    'None)[-1]'.format(attr_str, attr_assignment_str)
        exec('self.new_init = {}'.format(meth_str))
        #new_init = lambda a: a
        print(self.new_init)
        new_class_name = self.__class__.__name__ + 'ABC'
        child_class = type(new_class_name, (self.__class__, ), {'__init__': self.new_init})
        self.__class__ = child_class

    def _custom_hstack(self, Xs):
        """ concatenate dataframes if all branches are dataframes and it is possible
        """
        try:
            if False in [isinstance(x, pd.DataFrame) for x in Xs]:
                raise
            return pd.concat(Xs, axis=1)
        except:
            return self._custom_hstack(Xs)


    def _filter_skips(self, Xs):
        """ removes concatenated items if the transformer is marked to be skipped
        """
        lst = list(Xs)
        for i, attr in enumerate(self.skip_attr_list):
            if getattr(self, attr):
                del(lst[i])
        return tuple(lst)


    def set_params(self, **kwargs):
        # todo set the skip attributes
        super().set_params(**kwargs)

    def get_params(self, deep=True):
        params = super().get_params(deep=deep)
        for attr in self.skip_attr_list:
            params[attr] = getattr(self, attr)
        params['skip_attr_list'] = self.skip_attr_list
        return params

    def fit_transform(self, X, y=None, **fit_params):
        """Fit all transformers, transform the data and concatenate results.

        Parameters
        ----------
        X : iterable or array-like, depending on transformers
            Input data to be transformed.

        y : array-like of shape (n_samples, n_outputs), default=None
            Targets for supervised learning.

        Returns
        -------
        X_t : array-like or sparse matrix of \
                shape (n_samples, sum_n_components)
            hstack of results of transformers. sum_n_components is the
            sum of n_components (output dimension) over transformers.
        """


        results = self._parallel_func(X, y, fit_params, _fit_transform_one)
        if not results:
            # All transformers are None
            return np.zeros((X.shape[0], 0))

        Xs, transformers = zip(*results)
        self._update_transformer_list(transformers)

        Xs = self._filter_skips(Xs)
        return self._custom_hstack(Xs)

    def transform(self, X):
        """Transform X separately by each transformer, concatenate results.

        Parameters
        ----------
        X : iterable or array-like, depending on transformers
            Input data to be transformed.

        Returns
        -------
        X_t : array-like or sparse matrix of \
                shape (n_samples, sum_n_components)
            hstack of results of transformers. sum_n_components is the
            sum of n_components (output dimension) over transformers.
        """
        Xs = Parallel(n_jobs=self.n_jobs)(
            delayed(_transform_one)(trans, X, None, weight)
            for name, trans, weight in self._iter())
        if not Xs:
            # All transformers are None
            return np.zeros((X.shape[0], 0))

        Xs = self._filter_skips(Xs)
        return self._custom_hstack(Xs)



"""
Definition of pipes

"""




"""
Boolean 


"""
from pyadlml.constants import VALUE
from pyadlml.pipeline import Pipeline, TrainOnlyWrapper
from sklearn import RandomForestClassifier
from pyadlml.preprocessing import StateVectorEncoder, IndexEncoder, DropColumn, \
    LabelMatcher, DropTimeIndex, EventWindow, DropDuplicates



uci_torch_pipe = Pipeline([
    ('enc', IndexEncoder()),
    ('drop_obs', DropColumn(VALUE)),
    ('lbl', LabelMatcher()),
    ('drop_time_idx', DropTimeIndex()),
    ('time_windows', EventWindow()),
    ('passthrough', 'passthrough'),
])


uci_sklearn_iid_pipe = Pipeline([
    ('enc', StateVectorEncoder()),
    ('lbl', LabelMatcher()),
    ('drop_time_idx', DropTimeIndex()),
    ('drop_duplicates', TrainOnlyWrapper(DropDuplicates())),
    ('passthrough', 'passthrough'),
])



import zipfile
from pathlib import Path
from .local import _move_files_to_parent_folder
import urllib.request


class DatasetDownloader():
    def __init__(self,):
        pass


class WebZipDownloader(DatasetDownloader):


    def __init__(self, url, dataset_name):
        self.url = url
        self.dataset = dataset_name


    def download(self, dest: Path) -> None:
        """ Downloads zip file and moves the content into 
            the original folder.

        Parameters
        ----------
        dest : Path
            Path to 
        """
        fn_zip = self.url.split('/')[-1]
        fp_zip = Path(dest).joinpath(fn_zip)

        urllib.request.urlretrieve(self.url, fp_zip) 

        # Unzip data, remove and remove folder 
        with zipfile.ZipFile(fp_zip, "r") as zip_ref:
            zip_ref.extractall(dest)
        Path(fp_zip).unlink()
        files = list(dest.iterdir())
        if len(files) == 1:
            # Case when the zip's content is a folder
            _move_files_to_parent_folder(files[0])
        elif len(files) < 1:
            raise



class MegaDownloader(DatasetDownloader):
    def __init__(self, url, fn, url_cleaned, fn_cleaned):
        self.mega_url =  url
        self.fn = fn
        self.url_cleaned = url_cleaned
        self.fn_cleaned = fn_cleaned
    
    
    def download_cleaned(self, fp_dest: Path, ident=None) -> None:
        """ Download from mega
        dest:  Path
            Path to the cleaned file
        """
        fn_cleaned = self.fn_cleaned[ident] if ident is not None else self.fn_cleaned
        url_cleaned = self.url_cleaned[ident] if ident is not None else self.url_cleaned

        self._download_from_mega(fp_dest.parent, fn_cleaned, url_cleaned, unzip=False)
        import shutil
        src = fp_dest.parent.joinpath(fn_cleaned)
        sink = fp_dest
        shutil.move(src, sink)


    def download(self, dest: Path) -> None:
        self._download_from_mega(dest, self.fn, self.mega_url, unzip=True)

    def _download_from_mega(self, path_to_folder, file_name, url, unzip=True):
        """ Downloads dataset from MEGA and extracts it
        Parameters
        ----------
        path_to_folder : PosixPath or str
            The folder where the archive will be extracted to
        file_name : str
            The name of the file to be downloaded
        url : str
            The internet address where mega downloads the file
        unzip : bool, default=True

        """
        file_dp = Path(path_to_folder).joinpath(file_name)
        from mega import Mega

        # Download from mega
        m = Mega()    
        m.download_url(url, dest_path=str(path_to_folder), dest_filename=file_name)

        # Unzip data, remove
        if unzip:
            with zipfile.ZipFile(file_dp, "r") as zip_ref:
                zip_ref.extractall(path_to_folder)
            Path(file_dp).unlink()
            _move_files_to_parent_folder(path_to_folder.joinpath(file_name[:-4]))




from copy import copy
from pathlib import Path
from dash import Dash, dash_table, dcc, html
from dash.dependencies import Input, Output, State
import plotly.graph_objects as go
import pandas as pd
import argparse
import dash
from dash.exceptions import PreventUpdate
from pyadlml.constants import ACTIVITY, DEVICE, END_TIME, START_TIME, TIME, VALUE, ts2str, str2ts
from pyadlml.dataset._core.activities import ActivityDict, create_empty_activity_df
from pyadlml.dataset.act_assist import read_activities, write_activities, read_devices, write_devices
from pyadlml.dataset.plot.plotly.acts_and_devs import _determine_start_and_end, _plot_activities_into_fig
from pyadlml.dataset.plot.plotly.util import dash_get_trigger_element, dash_get_trigger_value, set_fig_range
import json
import numpy as np
from pyadlml.dataset.util import activity_order_by, device_order_by, df_difference, fetch_by_name, num_to_timestamp, select_timespan
from pyadlml.dataset.plot.plotly.activities import _set_compact_title, correction
from pyadlml.dataset.plot.plotly.util import deserialize_range, serialize_range, range_from_fig
import dash_bootstrap_components as dbc

"""
for dataset options look under pyadlml.constants:
    casas_aruba
    amsterdam
    mitlab_1
    mitlab_2
    aras
    kasteren_A
    kasteren_B
    kasteren_C
    tuebingen_2019
    uci_ordonezA
    uci_ordonezB
    act_assist
"""



DEL_KEY = 'DELETE'
JOIN_KEY = 'JOIN'


def json_to_df(df):
    try:
        return pd.read_json(df)
    except:
        return None
def df_to_json(df):
    try:
        return df.to_json(date_unit="ns")
    except:
        return ''

def read_code_and_update(fp_code, fp_code_dev, fp_df_acts, fp_df_devs):
    f = open(fp_code, mode='r')
    code_str = ''.join(f.readlines())
    f.close()

    f = open(fp_code_dev, mode='r')
    code_str_devs = ''.join(f.readlines())
    f.close()

    # TODO refactor, weird bug where locals()['df_acts'] is not overwritten
    #                when not expliciltly given as locals() dict and referenced with post exec command
    lcls = {'df_acts': read_activities(fp_df_acts)} 
    exec(code_str, {}, lcls)
    df_acts = lcls['df_acts']

    lcls = {'df_devs': read_devices(fp_df_devs)} 
    exec(code_str_devs, {}, lcls)
    df_devs = lcls['df_devs']
    return code_str, code_str_devs, df_acts, df_devs


def _get_room_y_pos(dev_order, dev2area, area_select):


    df = dev2area.copy()
    df[DEVICE] = df[DEVICE].astype("category")
    df[DEVICE] = df[DEVICE].cat.set_categories(dev_order)
    df = df.sort_values(by=DEVICE).reset_index(drop=True).reset_index()

    upper_y = len(dev_order) - df[df['area'] == area_select]['index'].iat[0]
    lower_y = len(dev_order)-1 - df[df['area'] == area_select]['index'].iat[-1]
    return lower_y, upper_y


def label_figure(df_acts, df_devs, dev2area, range_store=None, zoomed_range=[], states=False, area_select=''):
    if range_store is not None:
        st, et = range_store[0], range_store[1]
    else:
        st, et = None, None

    from pyadlml.dataset.plot.plotly.acts_and_devs import activities_and_devices
    dev_order = 'alphabetical' if dev2area is None else 'area'


    fig = activities_and_devices(df_devs, df_acts, dev2area=dev2area, zoomed_range=zoomed_range, states=states, dev_order=dev_order)

    if st:
        fig.add_vline(x=st, line_width=1, line_color="Grey")
    if et:
        fig.add_vline(x=et, line_width=1, line_color="Grey")

    if area_select:
        # Get y indicies of device bars/events
        dev_order = device_order_by(df_devs, rule='area', dev2area=dev2area)
        lower_y, upper_y = _get_room_y_pos(dev_order, dev2area, area_select)
        opacity = 0.5
    else:
        lower_y, upper_y, opacity = 1, 2, 0.0

    # Add highlight rectangle
    fig.add_hrect(
        y0=lower_y - 0.5, y1=upper_y - 0.5,
        fillcolor="LightSalmon", opacity=opacity,
        layer="below", line_width=0,
    )
    

    return fig



def reset_range_store(df_devs, df_acts):
    offset = pd.Timedelta('2s')
    st = min(df_devs.at[df_devs.index[0], TIME],
                df_acts.at[df_acts.index[0], START_TIME])
    et = max(df_devs.at[df_devs.index[-1], TIME],
                df_acts.at[df_acts.index[-1], END_TIME]
    )
    return [st - offset, et + offset]


if __name__ == '__main__':
    parser = argparse.ArgumentParser()
    parser.add_argument('-d', '--dataset', type=str, default='amsterdam')
    parser.add_argument('-i', '--identifier', type=str)
    parser.add_argument('-p', '--port', type=str, default='8050')
    parser.add_argument('-o', '--out-folder', type=str, default='/tmp/pyadlml/label_activities')
    parser.add_argument('-dh', '--data-home', type=str, default='/tmp/pyadlml/')
    args = parser.parse_args()

    out_folder = Path(args.out_folder)
    if not out_folder.exists():
        out_folder.mkdir(parents=True, exist_ok=True)

    tmp_folder = Path('/tmp/pyadlml/label_data/')
    tmp_folder.mkdir(exist_ok=True, parents=True)

    fp_df_acts = tmp_folder.joinpath('activities.csv')
    fp_df_devs = tmp_folder.joinpath('devices.csv')

    if fp_df_acts.exists():    
        fp_df_acts.unlink()
    
    if fp_df_devs.exists():    
        fp_df_devs.unlink()

    #fp_df_del = out_folder.joinpath('removed_activities.csv')
    #fp_df_add = out_folder.joinpath('new_activities.csv')
    fp_code = out_folder.joinpath('relabel_activities.py')
    fp_code_dev = out_folder.joinpath('relabel_devices.py')

    # Setup data
    data = fetch_by_name(args.dataset, identifier=args.identifier, \
                         cache=False, retain_corrections=True, keep_original=True)
    df_acts = data['activities']
    df_devs = data['devices']

    # TODO DEBUG
    try:
        dev2area = data['dev2area']
    except:
        dev2area = None
    
    activities = df_acts[ACTIVITY].unique()
    devices = df_devs[DEVICE].unique()

    app = Dash(__name__, external_stylesheets=[dbc.themes.BOOTSTRAP])

    #df_del = create_empty_activity_df()
    #df_add = create_empty_activity_df()
    df_acts = df_acts.sort_values(by=START_TIME).reset_index(drop=True)
    df_devs = df_devs.sort_values(by=TIME).reset_index(drop=True)

    write_activities(df_acts, fp_df_acts)
    write_devices(df_devs, fp_df_devs)

    #df_del.to_csv(fp_df_del, index=False)
    #df_add.to_csv(fp_df_add, index=False)

    code_str_acts = "# Awesome script\n" \
             + "import pandas as pd\n"\
             + "from pyadlml.constants import START_TIME, END_TIME, ACTIVITY\n"\
             + "from pyadlml.dataset._core.activities import get_index_matching_rows\n\n"

    code_str_devs = "# Awesome script\n" \
             + "import pandas as pd\n"\
             + "from pyadlml.constants import TIME, DEVICE, VALUE\n"\
             + "from pyadlml.dataset._core.devices import get_index_matching_rows\n\n"

    if not fp_code.exists():
        with open(fp_code, mode='w') as f:
            f.write(code_str_acts)

    if not fp_code_dev.exists():
        with open(fp_code_dev, mode='w') as f:
            f.write(code_str_devs)


    code_str, code_str_devs, df_acts, df_devs = read_code_and_update(fp_code, fp_code_dev, fp_df_acts, fp_df_devs)

    (a,b,_,_) = _determine_start_and_end(df_acts, df_devs, st=None, et=None)
    fig_range = a,b

    fig = label_figure(df_acts, df_devs, dev2area, range_store=fig_range)

    from pyadlml.dataset.plot.plotly.dashboard.layout import _build_range_slider

    app.layout = dbc.Container([
        dbc.Row(
            dcc.Graph(
                id='fig-activity',
                figure=fig,
            ), className='mb-3'
        ),
        dbc.Row([
            dbc.Col(
                dcc.RadioItems(['events', 'states'], 'events', id='and-vis'),
                md=4
            ),
            dbc.Col(
                dbc.Input(id='fig-range-input', placeholder='Paste range from dashboard'),
                md=4
            ),
            dbc.Col(
                dcc.Dropdown(id='area-select', options=[area for area in dev2area['area'].unique()], value=None),
                md=4
            ),

        ]),
        dbc.Row([
            _build_range_slider(df_acts, df_devs, a, b, b),
            dcc.Store(id='range-store', data=serialize_range(fig_range)),
            dcc.Store(id='range-slider-t-1', data=[0, 1]),       
        ],  className='mb-2'),
        dbc.Row([
            dbc.Col(md=5),
            dbc.Col(
                dbc.ButtonGroup([
                    dbc.Button(id="set-left", children="to left", color="secondary"), 
                    dbc.Button(id="set-frame", children="set frame", color="secondary"), 
                    dbc.Button(id="set-right", children="to right", color="secondary"), 
                ]),
            md=5),
            dbc.Col(md=2),
        ], className='mb-1', justify='center'),
        dbc.Row([
            dbc.Col(
                dash_table.DataTable(id='times', 
                                        columns=({'id': START_TIME, 'name': 'Start time'},
                                                {'id': END_TIME, 'name': 'End time'}),
                                        data=[{
                                        START_TIME: ts2str(df_acts.iloc[0,0]), 
                                        END_TIME: ts2str(df_acts.iloc[0,1])
                                        }],
                                        editable=True,
                                        style_cell={'font-size': '12px'}
                ), md=8
            ),
            dbc.Col(
                dbc.Row([
                    dbc.Col(
                        dbc.Row([
                            dbc.Select(id='activity',
                                        options=[ {"label": act, "value": act} for act in [*activities, DEL_KEY, JOIN_KEY]],
                            ),
                            dbc.Button(id="submit", children="Submit", color="primary")
                        ]), 
                    md=6),
                    dbc.Col(
                        dbc.Row([
                            dbc.Select(id='device',
                                        options=[ {"label": dev, "value": dev} for dev in devices],
                            ),
                            dcc.Dropdown(id='device-state',
                                        searchable=False, clearable=False, 
                                        options=[ {"label": cat, "value": cat} for cat in ['initial']],
                            ),
                            dbc.Button(id="submit-dev", children="Submit", color="primary")
                        ]),
                    md=6),
                ]),
            md=4,
            align='center', 
            className='mb-2')]
        ),
        dbc.Row(
            dbc.Textarea(id='comment', value='', readOnly=False, 
                        style={'height': '100px'}
                        #style={'resize': 'none', 'overflow':'hidden'}#, 'height': '100px'}
            ),
            style={'height':'100%'},
            className='mb-1',
        ),

    ], style={'width': 1800, 'margin': 'auto'})

    @app.callback(
        Output('device-state', 'options'),
        Output('device-state', 'value'),
        Input('device', 'value'),
    )
    def cat_select(value):
        trigger = dash_get_trigger_element()
        if trigger is None or trigger == '':
            raise PreventUpdate

        cats = df_devs.loc[df_devs[DEVICE] == value, VALUE].unique().tolist()

        if True in cats and False in cats and len(cats) == 2:
            cats = ['off', 'on']

        res = [{"label": str(cat), "value": cat} for cat in cats + ['del events']]

        return res, res[0]["label"]

    @app.callback(
        Output('fig-activity', 'figure'),
        Output('times', 'data'),
        Output('activity', 'value'),
        Output('range-store', 'data'),
        Output('range-slider-t-1', 'data'),
        Output('comment', 'value'),
        Input('and-vis', 'value'),
        Input('times', 'data'),
        Input('fig-activity', 'clickData'),
        Input('set-frame', 'n_clicks'),
        Input('set-left', 'n_clicks'),
        Input('set-right', 'n_clicks'),
        Input('submit', 'n_clicks'),
        Input('submit-dev', 'n_clicks'),
        Input('fig-range-input', 'value'),
        Input('area-select', 'value'),
        State('fig-activity', 'figure'),
        State('range-store', 'data'),
        State('range-slider-t-1', 'data'),
        State('activity', 'value'),
        State('device', 'value'),
        State('device-state', 'value'),
        State('comment', 'value'),
        State('range-slider', 'value'),
    )
    def display_output(and_vis, times, act_click, btn_set_frame, btn_left, btn_right, btn_submit, btn_submit_dev,
                       fig_range_input, area_select, fig, range_store, rs_tm1, activity, device, device_state, comment, range_select
                       #fig, range_store, rs_tm1, df_json, df_del_json, activity
    ):
        def from_01(y, r):
            mn, mx = r[0], r[1]
            """ [0,1] -> timestamps"""
            return y*(mx-mn)+mn 
        def to_01(x, r):
            """ timestamps -> [0,1]"""
            mn, mx = r[0], r[1]
            return (x-mn)/(mx-mn)


        trigger = dash_get_trigger_element()
        if trigger is None or trigger == '':
            raise PreventUpdate

        # Check if data has to be reloaded or callback is independent
        is_data_update = trigger not in ['set-frame', 'set-left', 'set-right', 'fig-range-input', 'area-select']
        is_figure_reload = trigger not in ['set-frame', 'set-left', 'set-right', 'fig-range-input', 'fig-activity', 'area-select']


        if is_data_update:
            code_str, code_str_devs, df_acts, df_devs = read_code_and_update(fp_code, fp_code_dev, fp_df_acts, fp_df_devs)

        fig_range = range_from_fig(fig)

        if range_store is not None:
            range_store = deserialize_range(range_store)
        else:
            range_store = reset_range_store(df_devs, df_acts)

        if trigger == 'set-frame':
            # Get range from outer boundaries
            range_store = copy(fig_range)

            # Zoom a little bit out
            offset = (fig_range[1] - fig_range[0])*0.05
            fig_range[0] = fig_range[0] - offset
            fig_range[1] = fig_range[1] + offset

            #rng[0], rng[1] = to_01(range_store[0], fig_range), to_01(range_store[1], fig_range)
        if trigger == 'set-left':
            range_store[0] = copy(fig_range[0])
            offset = (fig_range[1] - fig_range[0])*0.05
            fig_range[0] = fig_range[0] - offset
            #rng[0], rng[1] = to_01(range_store[0], fig_range), to_01(range_store[1], fig_range)
        if trigger == 'set-right':
            range_store[1] = copy(fig_range[1])
            offset = (fig_range[1] - fig_range[0])*0.05
            fig_range[1] = fig_range[1] + offset
            #rng[0], rng[1] = to_01(range_store[0], fig_range), to_01(range_store[1], fig_range)
            
        if trigger == 'submit-dev':
            data = {
                START_TIME: range_store[0], 
                END_TIME: range_store[1], 
                DEVICE:device,
                VALUE: device_state
            }
            if data[VALUE] in ['on', 'off']:
                data[VALUE] = {'on':True, 'off':False}[data[VALUE]]

            code_str_devs += "\n#" + "-"*40 + '\n'
            code_str_devs += '\n'.join(['# ' + s for s in comment.split('\n')]) + '\n\n'

            def del_op2str(row):
                if isinstance(row[VALUE], str):
                    row[VALUE] = "'" + row[VALUE] + "'"
                row_lst = f"[['{ts2str(row[TIME])}','{row[DEVICE]}',{row[VALUE]}]]"
                s = f"idx_to_del = get_index_matching_rows(df_devs, {row_lst})\n"
                s += "df_devs = df_devs.drop(index=idx_to_del).reset_index(drop=True)\n"
                return s


            def add_op2str(row):
                if isinstance(row[VALUE], str):
                    row[VALUE] = "'" + row[VALUE] + "'"
                s = "new_row=pd.Series(\n{"
                s += f"TIME: pd.Timestamp('{row[TIME]}'), DEVICE:'{row[DEVICE]}', VALUE:{row[VALUE]}"
                s += "})\ndf_devs = pd.concat([df_devs, new_row.to_frame().T], axis=0).sort_values(by=TIME).reset_index(drop=True)\n"
                return s

            def del_bulk_op2str(st, et, dev):
                s =  f"mask = (df_devs[DEVICE] == '{dev}')\\\n"
                s += f"     & ('{st}' < df_devs[TIME])\\\n"
                s += f"     & (df_devs[TIME] < '{et}')\n"
                s += f"df_devs = df_devs[~mask].sort_values(by=TIME).reset_index(drop=True)\n\n"
                return s

            df = df_devs[df_devs[DEVICE] == data[DEVICE]].copy().reset_index(drop=True)
            mask = (data[START_TIME] < df[TIME])\
                 & (df[TIME] < data[END_TIME])
            df_tmp = df[mask]

            if len(df_tmp) == len(df):
                print('Warning! Select a range before submitting.')
                raise PreventUpdate

            if not df_tmp.empty and data[VALUE]  == 'del events':
                code_str_devs += del_bulk_op2str(data[START_TIME], data[END_TIME], data[DEVICE])

            # Delete enveloped events and get index of preceeding event
            # and the last enveloped event
            elif not df_tmp.empty:
                for _, row in df_tmp.iterrows():
                    code_str_devs += del_op2str(row) + '\n'
                last_event = df_tmp.iloc[-1]
                prec_idx = df_tmp.index[0] - 1

                if last_event[VALUE] != data[VALUE]:
                    last_event[TIME] = data[END_TIME]
                    code_str_devs += add_op2str(last_event)

                if prec_idx >= 0:
                    prec_event = df.loc[prec_idx, ]

                if prec_idx < 0 or prec_event[VALUE] != data[VALUE]:
                    code_str_devs += add_op2str(
                        pd.Series({TIME:data[START_TIME], DEVICE:data[DEVICE], VALUE: data[VALUE]})
                    )
            else:
                prec_events = df[df[TIME] < data[START_TIME]]
                if not prec_events.empty:
                    prec_event = prec_events.iloc[-1]
                    prec_event[TIME] = data[END_TIME]
                    code_str_devs += add_op2str(prec_event)

                code_str_devs += add_op2str(
                    pd.Series({TIME:data[START_TIME], DEVICE:data[DEVICE], VALUE: data[VALUE]})
                )


            # Reset values

            range_store = reset_range_store(df_devs, df_acts)
            #rng = [0,1]
            comment = ''

            # Reload activity dataframe
            lcls = {'df_devs': read_devices(fp_df_devs)} 
            exec(code_str_devs, {}, lcls)
            df_devs = lcls['df_devs']


        elif trigger == 'submit':
            # Add activity to dataframe
            data = {
                START_TIME: range_store[0], 
                END_TIME: range_store[1], 
                ACTIVITY:activity
            }
            def del_op2str(row):
                row_lst = f"\t[['{ts2str(row[START_TIME])}','{ts2str(row[END_TIME])}','{row[ACTIVITY]}']],\n"
                return f"idx_to_del = get_index_matching_rows(df_acts, {row_lst})\n"\
                     + f"df_acts = df_acts.drop(index=idx_to_del).reset_index(drop=True)\n"

            def add_op2str(row):
                row_dict = "{"\
                        + f"START_TIME: pd.Timestamp('{(row[START_TIME])}'), "\
                        + f"END_TIME: pd.Timestamp('{row[END_TIME]}'), "\
                        + f"ACTIVITY: '{row[ACTIVITY]}'"\
                + "}"
                return f"new_row=pd.Series({row_dict})\n"\
                      + "df_acts = pd.concat([df_acts, new_row.to_frame().T], axis=0).reset_index(drop=True)\n"

            # Start adding to code string and add comment if given
            code_str += "\n#" + "-"*40 + '\n'
            code_str += '\n'.join(['# ' + s for s in comment.split('\n')]) + '\n\n'

            new_int = pd.Interval(data[START_TIME], data[END_TIME])
            df_acts[TIME] = df_acts[[START_TIME, END_TIME]].apply(lambda x: pd.Interval(x[START_TIME], x[END_TIME]), axis=1)
            mask = df_acts[TIME].apply(lambda x: x.overlaps(new_int))
            df_acts = df_acts[[START_TIME, END_TIME, ACTIVITY]]
            ov_act = df_acts[mask]

            if data[ACTIVITY] == DEL_KEY and not ov_act.empty:
                ov_act = df_acts[mask].iloc[0]
                #df_del = pd.concat([df_del, ov_act.to_frame().T], axis=0)
                code_str += '\n# Delete operation:\n' + del_op2str(ov_act)

                if data[START_TIME] < ov_act[START_TIME] and data[END_TIME] < ov_act[END_TIME]:
                    #    |~~~~~~~~|
                    # | DEL |
                    ov_act[START_TIME] = data[END_TIME]
                    #df_add = pd.concat([df_add, ov_act.to_frame().T], axis=0)
                    code_str += '\n' + add_op2str(ov_act)
                elif ov_act[START_TIME] < data[START_TIME] and data[END_TIME] < ov_act[END_TIME]:
                    #|~~~~~~~~|
                    # | DEL |
                    ov_actr = copy(ov_act)
                    ov_actl = copy(ov_act)
                    ov_actr[END_TIME] = data[START_TIME]
                    ov_actl[START_TIME] = data[END_TIME]
                    #df_add = pd.concat([df_add, ov_actr.to_frame().T, ov_actl.to_frame().T
                    #                   ], axis=0)

                    code_str += '\n' + add_op2str(ov_actr)
                    code_str += '\n' + add_op2str(ov_actl)
                elif ov_act[START_TIME] < data[START_TIME] and ov_act[END_TIME] < data[END_TIME]:
                    #|~~~~~~~~|
                    #     | DEL |
                    ov_act[END_TIME] = data[START_TIME]
                    #df_add = pd.concat([df_add, ov_act.to_frame().T], axis=0)
                    code_str += '\n' + add_op2str(ov_act)
                elif data[START_TIME] < ov_act[START_TIME] and ov_act[END_TIME] < data[END_TIME]: 
                    # |~~~~|
                    #|   DEL   |
                    pass
                elif ov_act[START_TIME] <= data[START_TIME] and data[END_TIME] <= ov_act[END_TIME]:
                    # |~~~~~|
                    # | DEL |
                    pass
                else:
                    raise NotImplementedError
            elif data[ACTIVITY] == JOIN_KEY and not ov_act.empty:
                assert len(ov_act) == 2, f'Select two activities to join. Only one selected: {ov_act}'
                ov_actr = ov_act.iloc[0]
                ov_actl = ov_act.iloc[1]
                code_str += '\n# Join operation: \n'
                code_str += del_op2str(ov_actr)
                code_str += del_op2str(ov_actl)
                ov_actr[END_TIME] = ov_actl[END_TIME]
                code_str += add_op2str(ov_actr)
            elif not ov_act.empty:
                assert len(ov_act) == 1, 'Can only edit one activity at a time.'
                ov_act = df_acts[mask].iloc[0]
                assert ov_act[ACTIVITY] == activity, 'Only select one activity'

                code_str += '\n# Modify operation:\n'

                #df_del = pd.concat([df_del, ov_act.to_frame().T], axis=0)
                code_str += '\n' + del_op2str(ov_act)

                # Case when an activity is modified 
                ov_act[START_TIME] = data[START_TIME]
                ov_act[END_TIME] = data[END_TIME]
                #df_add = pd.concat([df_add, ov_act.to_frame().T], axis=0)\
                #    .reset_index(drop=True)
                code_str += '\n' + add_op2str(ov_act)
            else:
                # Case when a new activity is being created
                assert data[ACTIVITY] != DEL_KEY, 'Can not add non-existing activity "DELETE"'
                act = pd.Series(data)
                code_str += '\n# Create operation:\n'

                #df_add = pd.concat([df_add, act.to_frame().T], axis=0)\
                #    .reset_index(drop=True)
                code_str += '\n' + add_op2str(act)

            # Reset values
            range_store = reset_range_store(df_devs, df_acts)
            #rng = [0,1]
            comment = ''

            # Reload activity dataframe
            lcls = {'df_acts': read_activities(fp_df_acts)} 
            exec(code_str, {}, lcls)
            df_acts = lcls['df_acts']

        elif trigger == 'times':
            range_store[0] = str2ts(times[0][START_TIME])
            range_store[1] = str2ts(times[0][END_TIME])

        #        # Set the left slider and then update left bar
        #        rng[1] = to_01(range_store[1], fig_range)
        #    else:
        #        raise PreventUpdate

        #elif trigger == 'update-range':
        #    # grey lines set slider 
        #    rng[0], rng[1] = to_01(range_store[0], fig_range), to_01(range_store[1], fig_range)
        
        
        elif trigger == 'fig-activity':
            if not 'Acts: ' in act_click['points'][0]['y']:
                raise PreventUpdate

            st = pd.Timestamp(act_click['points'][0]['base'])
            #et = str2ts(act_click['points'][0]['x'])
            act = df_acts[(df_acts[START_TIME]-pd.Timedelta('1us') < st)
                        & (st < df_acts[START_TIME]+pd.Timedelta('1us'))]
            # & (df_acts[END_TIME] == et)]
            if act.empty and len(act) == 1:
                raise PreventUpdate
            act = act.iloc[0]
            # Update ranges
            range_store = [act[START_TIME], act[END_TIME]]
            times[0][START_TIME] = ts2str(range_store[0])
            times[0][END_TIME] = ts2str(range_store[1])

            activity = act[ACTIVITY]

        elif trigger == 'fig-range-input':
            try:
                tmp = fig_range_input.split(',')
                fig_range = [str2ts(tmp[0]), str2ts(tmp[1])]
                fig = set_fig_range(fig, fig_range)
            except:
                raise PreventUpdate

        if trigger in ['set-frame', 'set-left', 'set-right', 'fig-activity']:
            # Update vlines without recomputing figure
            try:
                # Left vline
                fig['layout']['shapes'][0]['x0'] = str(range_store[0])
                fig['layout']['shapes'][0]['x1'] = str(range_store[0])

                # Right vline
                fig['layout']['shapes'][1]['x0'] = str(range_store[1]) 
                fig['layout']['shapes'][1]['x1'] = str(range_store[1])

                set_fig_range(fig, fig_range)
            except KeyError:
                trigger = 'llululul'

        if trigger == 'area-select':
            try:
                dev_order = copy(fig['layout']['yaxis']['categoryarray'])
                dev_order.reverse()
                y0, y1 = _get_room_y_pos(dev_order, dev2area, area_select)
                fig['layout']['shapes'][2]['y0'] = y0 - 0.5
                fig['layout']['shapes'][2]['y1'] = y1 - 0.5
                fig['layout']['shapes'][2]['opacity'] = 0.5
            except KeyError or IndexError:
                trigger = 'llululul'

        if is_figure_reload:
            import time
            tic = time.perf_counter()

            # Only view a subset of the activities
            dct_acts = ActivityDict.wrap(df_acts)
            start_time = min(df_devs[TIME].iloc[0], dct_acts.min_starttime())
            start_time = start_time.floor('D')
            end_time = max(df_devs[TIME].iloc[-1], dct_acts.max_endtime())
            end_time = end_time.ceil('D')

            st = num_to_timestamp(range_select[0], start_time=start_time, end_time=end_time)
            et = num_to_timestamp(range_select[1], start_time=start_time, end_time=end_time)
            curr_df_devs, curr_df_acts = select_timespan(df_acts=df_acts, df_devs=df_devs,
                                                        start_time=st, end_time=et, clip_activities=False)
            fig = label_figure(curr_df_acts, curr_df_devs, dev2area, range_store, fig_range, states=(and_vis == 'states'), area_select=area_select)
            toc = time.perf_counter()
            print(f"Updating figure in: {toc - tic:0.4f} seconds")


        # TODO, check if this works
        times[0][START_TIME] = ts2str(range_store[0])
        times[0][END_TIME] = ts2str(range_store[1])
        range_store = serialize_range(range_store)


        vars = [*locals().keys()]
        if 'code_str_devs' in vars:
            with open(fp_code_dev, mode='w') as f:
                f.write(code_str_devs)

        if 'code_str' in vars:
            with open(fp_code, mode='w') as f:
                f.write(code_str)

        return fig, times, activity, range_store, rs_tm1, comment

    app.run_server(debug=False, host='127.0.0.1', port=args.port)



from pathlib import Path
import pandas as pd

from pyadlml.constants import DEVICE, TIME, VALUE



def update_df(code, df, df_name):
    """ Update a dataframe 
    """
    if isinstance(code, Path):
        f = open(code, 'r')
        code = f.read()
        f.close()

    lcls = {df_name: df} 
    exec(code, {}, lcls)
    return lcls[df_name]


def remove_state(df_devs, device, state, state_cond, replacement=None):
    """ Remove a state
    """

    df = df_devs[(df_devs[DEVICE] == device)].copy()
    df['diff'] = df[TIME].shift(-1) - df[TIME]
    mask = df['diff'].apply(state_cond) & (df[VALUE] == state)
    idxs = df[mask | mask.shift(1)].index
    df_devs = df_devs.drop(idxs)

    if replacement is not None:
        new_state_length = pd.Timedelta(replacement[1])

        error_msg = 'Replacement went wrong. Given statelength is greater then states'
        error_msg += 'of at least one matching condition.'
        assert (df[mask]['diff'] >= new_state_length).all(), error_msg

        tmp = df[mask].copy()[[TIME, DEVICE, VALUE]]
        tmp[VALUE] = replacement[0]

        # Create offset events with values of succeeding state
        tmp2 = tmp.copy()
        tmp2[VALUE] = df[mask.shift(1) | False][VALUE].values
        tmp2[TIME] = tmp2[TIME] + new_state_length
        df_devs = pd.concat([df_devs, tmp, tmp2])

    return df_devs.reset_index(drop=True)\
                  .sort_values(by=TIME)